import os
import pathlib

from typing import (
    Collection,
    Generator,
    Mapping,
    Union,
    cast,
    List,
    NamedTuple,
    Optional,
    TypedDict,
)

import datasets
import jsonlines
import pytorch_lightning as pl
import torch
import torch.utils.data
import transformers

from datasets.fingerprint import Hasher
from loguru import logger
from torch.nn.utils.rnn import pad_sequence

# Nouvo plan : un lecteur de jsonlines custom qui prédécoupe en source/target avec attribut src et
# tgt, on charge ça dans dataset pour le système de cache, puis dans le dataloader on sample et
# quelque part dans le trainmodule on ajoute le bruit utiliser
# <https://huggingface.co/docs/datasets/loading#python-generator> comme ça on peut streamer l'entrée


class DataRow(TypedDict):
    source: str
    target: str
    src_lang: str
    tgt_lang: str


def extract_from_jsonline(
    example: Union[Mapping[str, str], Mapping[str, Mapping[str, str]]],
    langcode_sub: Mapping[str, str],
    denoise_langs: Collection[str],
    source_langs: Collection[str],
    target_langs: Collection[str],
) -> Generator[DataRow, None, None]:
    # We deal with both top-level tranlatifrdgggggggggggggggggggggggggggggggggggggggggwons and 🤗's conventional format for this task
    example = cast(Mapping[str, str], example.get("translation", example))
    for dns_lang in denoise_langs:
        if not (dns_str := example.get(dns_lang)):
            continue
        row: DataRow = {
            "source": dns_str,
            "target": dns_str,
            "src_lang": langcode_sub.get(dns_lang, dns_lang),
            "tgt_lang": langcode_sub.get(dns_lang, dns_lang),
        }
        yield row

    for src_lang in source_langs:
        if not (src_str := example.get(src_lang)):
            continue
        for tgt_lang in target_langs:
            if tgt_lang == src_lang:
                continue
            if not (tgt_str := example.get(tgt_lang)):
                continue
            row  = {
                "source": src_str,
                "target": tgt_str,
                "src_lang": langcode_sub.get(src_lang, src_lang),
                "tgt_lang": langcode_sub.get(tgt_lang, tgt_lang),
            }
            yield row


class EncodedSample(TypedDict):
    attention_mask: List[int]
    decoder_input_ids: List[int]
    input_ids: List[int]
    labels: List[int]
    src_lang: str
    src_text: str
    special_tokens_mask: List[int]
    tgt_lang: str
    tgt_text: str


# NOTE: this also caches the raw and encoded dataset in HF_DATASETS_CACHE, which is different from OUR cache
# end users can still manually set HF_DATASETS_CACHE if e.g. their home has a small quota
def encode_dataset(
    denoise_langs: Collection[str],
    langcode_sub: Mapping[str, str],
    save_path: pathlib.Path,
    source_langs: Collection[str],
    target_langs: Collection[str],
    text_path: Union[pathlib.Path, str],
    tokenizer: Union[transformers.PreTrainedTokenizer, transformers.PreTrainedTokenizerFast],
    tokenizer_name: str,
    max_length: Optional[int] = None,
):
    if not hasattr(tokenizer, "src_lang") or not hasattr(tokenizer, "tgt_lang"):
        raise ValueError(
            "The tokenizer used in mBART training must have `src_lang` and `tgt_lang` attributes."
        )

    logger.info(f"Loading data from {text_path}")

    with jsonlines.open(text_path) as in_stream:

        def gen():
            for example in in_stream:
                yield from extract_from_jsonline(
                    example=example,
                    langcode_sub=langcode_sub,
                    denoise_langs=denoise_langs,
                    source_langs=source_langs,
                    target_langs=target_langs,
                )

        raw_dataset = datasets.Dataset.from_generator(gen)

    logger.info("Preprocessing dataset")

    def preprocess(example: DataRow) -> EncodedSample:
        # FIXME: the following lines seem to bug with extra added langs
        tokenizer.src_lang = example["src_lang"]  # type: ignore
        tokenizer.tgt_lang = example["tgt_lang"]  # type: ignore
        tokenized = tokenizer(
            example["source"],
            text_target=example["target"],
            add_special_tokens=True,
            max_length=max_length,
            return_attention_mask=True,
            return_special_tokens_mask=True,
            truncation=True,
        )
        # NOTE(2023-02-12): This stuff with decoder_input_ids is what 🤗 transformers's
        # `prepare_decoder_input_ids_from_labels`/[`shift_tokens_right`](https://github.com/huggingface/transformers/blob/c836f77266be9ace47bff472f63caf71c0d11333/src/transformers/models/mbart/modeling_mbart.py#L62)
        # does for mBART, but it seems a bit weird See also 🤗 transformers issue
        # [#19500](https://github.com/huggingface/transformers/issues/19500).
        labels = cast(List[int], tokenized["labels"])
        decoder_input_ids = [labels[-1], *labels[:-1]]
        # No loss for the langid token
        labels[0] = -100
        return EncodedSample(
            attention_mask=tokenized.attention_mask,
            decoder_input_ids=decoder_input_ids,
            input_ids=tokenized.input_ids,
            labels=labels,
            src_lang=example["src_lang"],
            src_text=example["source"],
            special_tokens_mask=cast(List[int], tokenized["special_tokens_mask"]),
            tgt_lang=example["tgt_lang"],
            tgt_text=example["target"],
        )

    raw_fingerprint = raw_dataset._fingerprint  # type: ignore
    encoded_dataset = cast(datasets.Dataset, raw_dataset).map(
        preprocess,
        batched=False,
        desc="Tokenizing",
        remove_columns=["source", "target"],
        new_fingerprint=Hasher.hash(
            f"{raw_fingerprint}-{tokenizer_name}-{source_langs}×{target_langs}×{denoise_langs}-{max_length}"
        ),
    )
    logger.info(f"Saving dataset to {save_path}")
    # FIXME: this causes an obscure crash when two instances want to access the same --cache-dir
    encoded_dataset.save_to_disk(str(save_path))


class MBARTBatch(NamedTuple):
    attention_mask: torch.Tensor
    decoder_input_ids: torch.Tensor
    input_ids: torch.Tensor
    labels: torch.Tensor
    src_lang: List[str]
    src_text: List[str]
    special_tokens_mask: torch.Tensor
    tgt_lang: List[str]
    tgt_text: List[str]


class TwoMBartBatches(NamedTuple):
    denoise: Optional[MBARTBatch]
    translate: Optional[MBARTBatch]


class MBartLoader(torch.utils.data.DataLoader[EncodedSample]):
    def __init__(
        self,
        dataset: torch.utils.data.Dataset[EncodedSample],
        tokenizer: Union[transformers.PreTrainedTokenizer, transformers.PreTrainedTokenizerFast],
        *args,
        **kwargs,
    ):
        self.dataset: torch.utils.data.Dataset[EncodedSample]
        if "collate_fn" not in kwargs:
            kwargs["collate_fn"] = self.collate
        super().__init__(dataset, *args, **kwargs)
        self.tokenizer = tokenizer
        padding_value = self.tokenizer.pad_token_id
        if padding_value is None:
            raise ValueError("Tokenizers without a padding id are not supported")
        self._padding_value = padding_value

    def collate(self, batch: List[EncodedSample]) -> TwoMBartBatches:
        # NOTE(2021-08-12): we have to pad/batch manually instead of deferring to 🤗, since the fast
        # tokenizers can't take pre-encoded inputs (yet?)
        padded_input_ids = pad_sequence(
            [torch.tensor(sample["input_ids"], dtype=torch.long) for sample in batch],
            batch_first=True,
            padding_value=self._padding_value,
        )
        padded_decoder_input_ids = pad_sequence(
            [torch.tensor(sample["decoder_input_ids"], dtype=torch.long) for sample in batch],
            batch_first=True,
            padding_value=self._padding_value,
        )
        padded_labels = pad_sequence(
            [torch.tensor(sample["labels"], dtype=torch.long) for sample in batch],
            batch_first=True,
            padding_value=-100,
        )
        special_tokens_mask = pad_sequence(
            [torch.tensor(sample["special_tokens_mask"], dtype=torch.long) for sample in batch],
            batch_first=True,
            padding_value=0,
        )
        attention_mask = padded_input_ids.ne(self._padding_value)

        # NOTE(2023-02-15): doing it this way probably results in overpadding at some point
        denoise_indices = [i for i, s in enumerate(batch) if s["src_lang"] == s["tgt_lang"]]
        if denoise_indices:
            denoise_batch = MBARTBatch(
                attention_mask=attention_mask[denoise_indices],
                decoder_input_ids=padded_decoder_input_ids[denoise_indices],
                input_ids=padded_input_ids[denoise_indices],
                labels=padded_labels[denoise_indices],
                src_lang=[batch[i]["src_lang"] for i in denoise_indices],
                src_text=[batch[i]["src_text"] for i in denoise_indices],
                special_tokens_mask=special_tokens_mask[denoise_indices],
                tgt_lang=[batch[i]["tgt_lang"] for i in denoise_indices],
                tgt_text=[batch[i]["tgt_text"] for i in denoise_indices],
            )
        else:
            denoise_batch = None

        translate_indices = [i for i, s in enumerate(batch) if s["src_lang"] != s["tgt_lang"]]
        if translate_indices:
            translate_batch = MBARTBatch(
                attention_mask=attention_mask[translate_indices],
                decoder_input_ids=padded_decoder_input_ids[translate_indices],
                input_ids=padded_input_ids[translate_indices],
                labels=padded_labels[translate_indices],
                src_lang=[batch[i]["src_lang"] for i in translate_indices],
                src_text=[batch[i]["src_text"] for i in translate_indices],
                special_tokens_mask=special_tokens_mask[translate_indices],
                tgt_lang=[batch[i]["tgt_lang"] for i in translate_indices],
                tgt_text=[batch[i]["tgt_text"] for i in translate_indices],
            )
        else:
            translate_batch = None
        return TwoMBartBatches(denoise=denoise_batch, translate=translate_batch)


class MBartDataModule(pl.LightningDataModule):
    def __init__(
        self,
        denoise_langs: Collection[str],
        langcode_sub: Mapping[str, str],
        loader_batch_size: int,
        num_workers: int,
        source_langs: Collection[str],
        target_langs: Collection[str],
        tokenizer: Union[transformers.PreTrainedTokenizer, transformers.PreTrainedTokenizerFast],
        tokenizer_name: str,
        train_path: Union[str, pathlib.Path],
        data_dir: Optional[pathlib.Path] = None,
        max_length: Optional[int] = None,
        val_path: Optional[Union[str, pathlib.Path]] = None,
    ):
        super().__init__()
        self.denoise_langs = sorted(set(denoise_langs))
        self.langcode_sub = langcode_sub
        self.loader_batch_size = loader_batch_size
        self.max_length = max_length
        self.num_workers = num_workers
        self.source_langs = sorted(set(source_langs))
        self.target_langs = sorted(set(target_langs))
        self.tokenizer = tokenizer
        self.tokenizer_name = tokenizer_name
        self.train_path = train_path
        self.val_path = val_path

        if data_dir is None:
            self.data_dir = pathlib.Path(self.train_path).parent
            if not self.data_dir.exists():
                raise ValueError(
                    "You must provide a cache path if you are loading a dataset from an url."
                )
        else:
            self.data_dir = data_dir

        self.train_dataset_path = self.data_dir / "train_set"
        self.train_dataset_path.mkdir(exist_ok=True, parents=True)

        self.val_dataset_path: Optional[pathlib.Path]
        if self.val_path is not None:
            self.val_dataset_path = self.data_dir / "val_set"
            self.val_dataset_path.mkdir(exist_ok=True, parents=True)
        else:
            self.val_dataset_path = None

        self.train_dataset = None
        self.val_dataset = None

    def prepare_data(self):
        # NOTE (2021-08-12): This should'nt be needed since this method should only be called on rank 0, but since it
        # is called in every process AND before DDP init (at least in SLURM) we have to enforce it
        # ourselves
        # TODO (2023-01-07): see if the note above is still true
        if os.environ.get("SLURM_PROCID", "0") == "0":
            encode_dataset(
                denoise_langs=self.denoise_langs,
                langcode_sub=self.langcode_sub,
                max_length=self.max_length,
                save_path=self.train_dataset_path,
                source_langs=self.source_langs,
                target_langs=self.target_langs,
                text_path=self.train_path,
                tokenizer=self.tokenizer,
                tokenizer_name=self.tokenizer_name,
            )
            if self.val_path is not None:
                assert self.val_dataset_path is not None
                encode_dataset(
                    denoise_langs=self.denoise_langs,
                    langcode_sub=self.langcode_sub,
                    max_length=self.max_length,
                    save_path=self.val_dataset_path,
                    source_langs=self.source_langs,
                    target_langs=self.target_langs,
                    text_path=self.train_path,
                    tokenizer=self.tokenizer,
                    tokenizer_name=self.tokenizer_name,
                )

    def setup(self, stage=None):
        self.train_dataset = cast(
            datasets.Dataset, datasets.load_from_disk(str(self.train_dataset_path))
        )
        if self.val_dataset_path is not None:
            self.val_dataset = cast(
                datasets.Dataset, datasets.load_from_disk(str(self.val_dataset_path))
            )

    def train_dataloader(self):
        if self.train_dataset is None:
            return None
        # FIXME(2023-02-07): that cast hereunder is wrong, self.train_dataset is **not** a torch Dataset
        return MBartLoader(
            cast(torch.utils.data.Dataset[EncodedSample], self.train_dataset),
            batch_size=self.loader_batch_size,
            num_workers=self.num_workers,
            shuffle=True,
            tokenizer=self.tokenizer,
        )

    def val_dataloader(self):
        if self.val_dataset is None:
            return None
        # FIXME(2023-02-07): that cast hereunder is wrong, self.val_dataset is **not** a torch Dataset
        return MBartLoader(
            cast(torch.utils.data.Dataset[EncodedSample], self.val_dataset),
            batch_size=self.loader_batch_size,
            num_workers=self.num_workers,
            shuffle=False,
            tokenizer=self.tokenizer,
        )



import pathlib
from typing import Dict, Optional

import pytorch_lightning as pl
import torch
import transformers
from loguru import logger
from pytorch_lightning.utilities.rank_zero import rank_zero_only


def get_internal_transformer_model(
    model: transformers.PreTrainedModel,
) -> torch.nn.Module:
    # There is no consensus in hf transformers as to how the underlying transformer of a MLM
    # model is called
    transformer_model = next(
        tr_model
        for transformer_name in ("bert", "electra", "roberta", "transformer")
        for tr_model in [getattr(model, transformer_name, None)]
        if tr_model is not None
    )
    return transformer_model


def reset_transformer_vocab(model: transformers.PreTrainedModel):
    logger.info("Reinitializing model embeddings")
    transformer_model = get_internal_transformer_model(model)
    if isinstance(transformer_model.embeddings, torch.nn.Embedding):
        transformer_model.embeddings.reset_parameters()
    # Assume a custom huggingface embedding class
    else:
        transformer_model.embeddings = type(transformer_model.embeddings)(transformer_model.config)
    logger.info("Reinitializing LM head")
    # There is no consensus in hf transformers as to how the LM head of a MLM model is
    # called so we have to do an ugly song and dance here
    lm_head_name = next(
        layer_name for layer_name in ("lm_head", "cls", "pred_layer") if hasattr(model, layer_name)
    )
    setattr(model, lm_head_name, type(getattr(model, lm_head_name))(model.config))


@rank_zero_only
def save_model(
    model: transformers.PreTrainedModel,
    save_dir: pathlib.Path,
    tokenizer: Optional[transformers.PreTrainedTokenizer] = None,
):
    """Save a transformer model."""
    save_dir.mkdir(parents=True, exist_ok=True)
    logger.info(f"Saving model to {save_dir}")
    model.save_pretrained(str(save_dir))
    if tokenizer is not None:
        logger.info(f"Saving tokenizer to {save_dir}")
        tokenizer.save_pretrained(str(save_dir), legacy_format=not tokenizer.is_fast)


class ShareTransformersEmbeddingsCallback(pl.Callback):
    def __init__(
        self,
        leader: transformers.PreTrainedModel,
        follower: transformers.PreTrainedModel,
    ):
        self.leader = leader
        self.follower = follower
        self.leader_transformer = get_internal_transformer_model(leader)
        self.follower_transformer = get_internal_transformer_model(follower)
        if not hasattr(self.leader_transformer, "embeddings") or not isinstance(
            self.leader_transformer, torch.nn.Module
        ):
            raise ValueError(
                f"Unsupported transformer: {self.leader_transformer} has no embedding submodule"
            )
        if not hasattr(self.follower_transformer, "embeddings") or not isinstance(
            self.follower_transformer, torch.nn.Module
        ):
            raise ValueError(
                f"Unsupported transformer: {self.follower_transformer} has no embedding submodule"
            )
        if not isinstance(
            self.leader_transformer.embeddings,
            type(self.follower_transformer.embeddings),
        ):
            logger.warning(
                "Sharing embeddings between different model types might not work well"
                f" {type(self.follower_transformer.embeddings)} vs {type(self.leader_transformer.embeddings)}"
            )

    def on_train_start(self, trainer, pl_module):
        self.follower_transformer.embeddings = self.leader_transformer.embeddings
        # Tying weights if needed
        # This works because in 🤗, embeddings tying makes output embeddings follow input embeddings
        # so here everyone will transitively follow the leader's embeddings
        if hasattr(self.follower, "tie_weights"):
            self.follower.tie_weights()

    def on_train_end(self, trainer, pl_module):
        self.follower_transformer.embeddings = type(self.follower_transformer.embeddings)(
            self.follower_transformer.config
        )
        self.follower_transformer.embeddings.load_state_dict(
            self.leader_transformer.embeddings.state_dict()
        )
        # Retying weights, here again, we rely on the fact that this uses
        # the input embeddings as ground truth
        if hasattr(self.follower, "tie_weights"):
            self.follower.tie_weights()


class CombiningLayer(torch.nn.Module):
    def __init__(self, leader: torch.nn.Module, follower: torch.nn.Module):
        super().__init__()
        self.leader = leader
        self.follower = follower

    def forward(self, *args, **kwargs):
        with torch.no_grad():
            leader_out = self.leader(*args, **kwargs)
        follower_out = self.follower(*args, **kwargs)

        return leader_out + follower_out


class OneWayShareTransformersEmbeddingsCallback(pl.Callback):
    def __init__(
        self,
        leader: transformers.PreTrainedModel,
        follower: transformers.PreTrainedModel,
    ):
        self.leader = leader
        self.follower = follower
        self.leader_transformer = get_internal_transformer_model(leader)
        self.follower_transformer = get_internal_transformer_model(follower)
        if not hasattr(self.leader_transformer, "embeddings") or not isinstance(
            self.leader_transformer, torch.nn.Module
        ):
            raise ValueError(
                f"Unsupported transformer: {self.leader_transformer} has no embedding submodule"
            )
        if not hasattr(self.follower_transformer, "embeddings") or not isinstance(
            self.follower_transformer, torch.nn.Module
        ):
            raise ValueError(
                f"Unsupported transformer: {self.follower_transformer} has no embedding submodule"
            )
        if not isinstance(
            self.leader_transformer.embeddings,
            type(self.follower_transformer.embeddings),
        ):
            logger.warning(
                "Sharing embeddings between different model types might not work well:"
                f" {type(self.follower_transformer.embeddings)} vs {type(self.leader_transformer.embeddings)}"
            )

        self.replaced_layer: Dict[str, CombiningLayer] = dict()

    def on_train_start(self, trainer, pl_module):
        for layer_name, layer in self.follower.embeddings.named_children():
            if isinstance(layer, torch.nn.Embedding):
                logger.debug(f"One-way sharing of {layer_name}.")
                torch.nn.init.zeros_(layer)
                leader_equiv = getattr(self.leader.embeddings, layer_name)
                combiner = CombiningLayer(leader=leader_equiv, follower=layer)
                setattr(self.follower.embeddings, layer_name, combiner)
                self.replaced_layer[layer_name] = combiner

        if not self.replaced_layer:
            logger.warning("No embeddings actually shared, you should probably check your config.")

    def on_train_end(self, trainer, pl_module):
        for layer_name, combiner in self.replaced_layer.items():
            combiner.follower.weight += combiner.leader.weight
            setattr(self.follower.embeddings, layer_name, combiner.follower)
        # Retying weights, here again, we rely on the fact that this uses
        # the input embeddings as ground truth
        if hasattr(self.follower, "tie_weights"):
            self.follower.tie_weights()
        self.replaced_layer = dict()


class TieEmbeddingsCallback(pl.Callback):
    def __init__(
        self,
        leader_embeddings: torch.nn.Embedding,
        follower_embeddings: torch.nn.Embedding,
    ):
        if leader_embeddings.weight.shape != follower_embeddings.weight.shape:
            raise ValueError(
                "Embeddings must have the same shape to be tied:"
                " {leader_embeddings.weight.shape} vs {follower_embeddings.weight.shape}"
            )
        self.leader_embeddings = leader_embeddings
        self.follower_embeddings = follower_embeddings

    def on_train_start(self, trainer, pl_module):
        self.follower_embeddings.weight = self.leader_embeddings.weight

    def on_train_end(self, trainer, pl_module):
        self.follower_embeddings.weight = self.follower_embeddings.weight.detach().clone()



import os
import pathlib

from typing import Union, cast, List, NamedTuple, Optional, Sequence, TypedDict

import datasets
import pytorch_lightning as pl
import torch
import torch.utils.data
import transformers

from datasets.fingerprint import Hasher
from loguru import logger
from torch.nn.utils.rnn import pad_sequence


# NOTE: this also caches the raw and encoded dataset in HF_DATASETS_CACHE, which is different from OUR cache
# end users can still manually set HF_DATASETS_CACHE if e.g. their home has a small quota
def encode_dataset(
    save_path: pathlib.Path,
    text_path: Union[pathlib.Path, str],
    tokenizer: Union[transformers.PreTrainedTokenizer, transformers.PreTrainedTokenizerFast],
    tokenizer_name: str,
    max_length: Optional[int] = None,
):
    logger.info(f"Loading data from {text_path}")
    try:
        full_dataset = datasets.load_dataset("text", data_files=str(text_path), split="train")
    except FileNotFoundError as e:
        if isinstance(text_path, str):
            dataset_name, dataset_config, dataset_split = text_path.split(":")
            full_dataset = datasets.load_dataset(
                dataset_name,
                name=dataset_config if dataset_config else None,
                split=dataset_split if dataset_split else None,
            )
        else:
            raise e

    # FIXME: why don't we use the default fingerprinting here?
    raw_dataset = cast(datasets.Dataset, full_dataset).filter(
        (lambda example: len(example) > 0 and not example.isspace()),
        input_columns="text",
        new_fingerprint=Hasher.hash(f"{full_dataset._fingerprint}-noempty"),
    )
    logger.info("Tokenizing")
    encoded_dataset = raw_dataset.map(
        (
            lambda examples: tokenizer(
                examples,
                add_special_tokens=True,
                max_length=max_length,
                return_special_tokens_mask=True,
                truncation=True,
            )
        ),
        batched=True,
        desc="Tokenizing",
        input_columns="text",
        remove_columns=["text"],
        new_fingerprint=Hasher.hash(f"{raw_dataset._fingerprint}-{tokenizer_name}-{max_length}"),
    )
    logger.info(f"Saving dataset to {save_path}")
    # FIXME: this causes an obscure crash when two instance want to access the same --cache-dir
    encoded_dataset.save_to_disk(str(save_path))


class TextBatch(NamedTuple):
    """A batch of text for self-supervised tasks.

    ## Attributes

    - `tokens` A batch of encoded (with special tokens) and padded tokens.
    - `attention_mask` A boolean mask, `True` for content and special tokens, `False` for padding.
    - `internal_tokens_mask` A boolean mask, `True` for content tokens, `False` for padding and
      special tokens.
    - `token_type_ids` The `token_type_ids` tensor needed internally for hugginface transformers
      implementations.
    """

    tokens: torch.Tensor
    attention_mask: torch.Tensor
    internal_tokens_mask: torch.Tensor
    token_type_ids: torch.Tensor


class EncodedSample(TypedDict):
    attention_mask: List[int]
    input_ids: List[int]
    special_tokens_mask: List[int]


class TextLoader(torch.utils.data.DataLoader[EncodedSample]):
    def __init__(
        self,
        dataset: torch.utils.data.Dataset[EncodedSample],
        tokenizer: Union[transformers.PreTrainedTokenizer, transformers.PreTrainedTokenizerFast],
        *args,
        **kwargs,
    ):
        self.dataset: torch.utils.data.Dataset[EncodedSample]
        if "collate_fn" not in kwargs:
            kwargs["collate_fn"] = self.collate
        super().__init__(dataset, *args, **kwargs)
        self.tokenizer = tokenizer
        padding_value = self.tokenizer.pad_token_id
        if padding_value is None:
            raise ValueError("Tokenizers without a padding id are not supported")
        self._padding_value = padding_value

    def collate(self, batch: Sequence[EncodedSample]) -> TextBatch:
        # FIXME(2023-02-11): the note below might have become inaccurate
        # NOTE: we have to pad/batch manually instead of deferring to huggingface, since the fast
        # tokenizers can't take pre-encoded inputs (yet?)
        padded_batch = pad_sequence(
            [torch.tensor(sample["input_ids"], dtype=torch.long) for sample in batch],
            batch_first=True,
            padding_value=self._padding_value,
        )
        padding_mask = padded_batch.eq(self._padding_value)
        # FIXME: Is the next line general enough?
        token_type_ids = torch.zeros_like(padded_batch)
        # We only deal with single sequences here
        attention_mask = padding_mask.logical_not()

        special_tokens_mask = pad_sequence(
            [torch.tensor(sample["special_tokens_mask"], dtype=torch.bool) for sample in batch],
            batch_first=True,
            padding_value=False,
        )
        internal_tokens_mask = special_tokens_mask.logical_or(padding_mask)
        return TextBatch(
            tokens=padded_batch,
            attention_mask=attention_mask,
            internal_tokens_mask=internal_tokens_mask,
            token_type_ids=token_type_ids,
        )


class TextDataModule(pl.LightningDataModule):
    def __init__(
        self,
        loader_batch_size: int,
        num_workers: int,
        tokenizer: Union[transformers.PreTrainedTokenizer, transformers.PreTrainedTokenizerFast],
        tokenizer_name: str,
        train_path: Union[str, pathlib.Path],
        data_dir: Optional[pathlib.Path] = None,
        max_length: Optional[int] = None,
        val_path: Optional[Union[str, pathlib.Path]] = None,
    ):
        super().__init__()
        self.loader_batch_size = loader_batch_size
        self.max_length = max_length
        self.num_workers = num_workers
        self.tokenizer = tokenizer
        self.tokenizer_name = tokenizer_name
        self.train_text = train_path
        self.val_text = val_path

        if data_dir is None:
            self.data_dir = pathlib.Path(self.train_text).parent
            if not self.data_dir.exists():
                raise ValueError(
                    "You must provide a cache path if you are loading a dataset from an url."
                )
        else:
            self.data_dir = data_dir

        self.train_dataset_path = self.data_dir / "train_set"
        self.train_dataset_path.mkdir(exist_ok=True, parents=True)

        self.val_dataset_path: Optional[pathlib.Path]
        if self.val_text is not None:
            self.val_dataset_path = self.data_dir / "val_set"
            self.val_dataset_path.mkdir(exist_ok=True, parents=True)
        else:
            self.val_dataset_path = None

        self.train_dataset = None
        self.val_dataset = None

    def prepare_data(self):
        # NOTE(2021-08-12): This should'nt be needed since this method should only be called on rank 0, but since it
        # is called in every process AND before DDP init (at least in SLURM) we have to enforce it
        # ourselves
        # TODO (2023-01-07): see if the note above is still true
        if os.environ.get("SLURM_PROCID", "0") == "0":
            encode_dataset(
                max_length=self.max_length,
                save_path=self.train_dataset_path,
                text_path=self.train_text,
                tokenizer=self.tokenizer,
                tokenizer_name=self.tokenizer_name,
            )
            if self.val_text is not None:
                assert self.val_dataset_path is not None
                encode_dataset(
                    max_length=self.max_length,
                    save_path=self.val_dataset_path,
                    text_path=self.val_text,
                    tokenizer=self.tokenizer,
                    tokenizer_name=self.tokenizer_name,
                )

    def setup(self, stage=None):
        self.train_dataset = datasets.load_from_disk(str(self.train_dataset_path))
        if self.val_dataset_path is not None:
            self.val_dataset = datasets.load_from_disk(str(self.val_dataset_path))

    def train_dataloader(self):
        if self.train_dataset is None:
            return None
        # FIXME(2023-02-07): that cast hereunder is wrong, self.train_dataset is **not** a torch Dataset
        return TextLoader(
            cast(torch.utils.data.Dataset[EncodedSample], self.train_dataset),
            batch_size=self.loader_batch_size,
            num_workers=self.num_workers,
            shuffle=True,
            tokenizer=self.tokenizer,
        )

    def val_dataloader(self):
        if self.val_dataset is None:
            return None
        # FIXME(2023-02-07): that cast hereunder is wrong, self.val_dataset is **not** a torch Dataset
        return TextLoader(
            cast(torch.utils.data.Dataset[EncodedSample], self.val_dataset),
            batch_size=self.loader_batch_size,
            num_workers=self.num_workers,
            shuffle=False,
            tokenizer=self.tokenizer,
        )



import os
import pathlib
from typing import List, Optional, Tuple, Union

import pytest
import pytest_console_scripts
import torch.cuda

accelerators_strategies_devices = [
    ("cpu", None, None),
    ("cpu", "ddp_spawn", 2),
    ("cpu", "ddp_spawn_find_unused_parameters_false", 2),
]
if torch.cuda.is_available():
    accelerators_strategies_devices.append(("gpu", None, None))
    if torch.cuda.device_count() > 1:
        accelerators_strategies_devices.append(("gpu", "ddp_spawn", 2))
        accelerators_strategies_devices.append(("gpu", "ddp_spawn_find_unused_parameters_false", 2))
        # accelerators_strategies_devices.append(("gpu", "fsdp_native", 2))


def test_train_tokenizer(
    raw_text_path: pathlib.Path,
    script_runner: pytest_console_scripts.ScriptRunner,
    tmp_path: pathlib.Path,
):
    ret = script_runner.run(
        "zeldarose-tokenizer",
        "--vocab-size",
        "4096",
        "--out-path",
        str(tmp_path / "tokenizer"),
        "--model-name",
        "my-muppet",
        str(raw_text_path),
    )
    assert ret.success


@pytest.mark.parametrize(
    "accelerators_strategies_devices",
    [pytest.param(v, id="+".join(map(str, v))) for v in accelerators_strategies_devices],
)
def test_train_mbart(
    accelerators_strategies_devices: Tuple[str, Optional[str], Optional[int]],
    mbart_model_config: Union[pathlib.Path, str],
    mbart_task_config: pathlib.Path,
    translation_dataset_path: pathlib.Path,
    script_runner: pytest_console_scripts.ScriptRunner,
    tmp_path: pathlib.Path,
):
    accelerator, strategy, devices = accelerators_strategies_devices
    extra_args: List[str] = []
    if strategy is not None:
        extra_args.extend(["--strategy", strategy])
    if devices is not None:
        extra_args.extend(["--num-devices", str(devices)])

    ret = script_runner.run(
        "zeldarose-transformer",
        "--accelerator",
        accelerator,
        "--config",
        str(mbart_task_config),
        "--tokenizer",
        str(mbart_model_config),
        "--model-config",
        str(mbart_model_config),
        "--device-batch-size",
        "2",
        "--out-dir",
        str(tmp_path / "train-out"),
        "--cache-dir",
        str(tmp_path / "cache"),
        "--val-text",
        str(translation_dataset_path),
        str(translation_dataset_path),
        "--max-epochs",
        "2",
        *extra_args,
        env={"TORCH_DISTRIBUTED_DEBUG": "DETAIL", **os.environ},
    )
    assert ret.success


@pytest.mark.parametrize(
    "accelerators_strategies_devices",
    [pytest.param(v, id="+".join(map(str, v))) for v in accelerators_strategies_devices],
)
def test_train_mlm(
    accelerators_strategies_devices: Tuple[str, Optional[str], Optional[int]],
    mlm_model_config: Union[pathlib.Path, str],
    mlm_task_config: pathlib.Path,
    raw_text_path: pathlib.Path,
    script_runner: pytest_console_scripts.ScriptRunner,
    tmp_path: pathlib.Path,
    tokenizer_name_or_path: Union[pathlib.Path, str],
):
    accelerator, strategy, devices = accelerators_strategies_devices
    extra_args: List[str] = []
    if strategy is not None:
        extra_args.extend(["--strategy", strategy])
    if devices is not None:
        extra_args.extend(["--num-devices", str(devices)])

    ret = script_runner.run(
        "zeldarose-transformer",
        "--accelerator",
        accelerator,
        "--config",
        str(mlm_task_config),
        "--tokenizer",
        str(tokenizer_name_or_path),
        "--model-config",
        str(mlm_model_config),
        "--device-batch-size",
        "8",
        "--out-dir",
        str(tmp_path / "train-out"),
        "--cache-dir",
        str(tmp_path / "tokenizer-cache"),
        "--val-text",
        str(raw_text_path),
        str(raw_text_path),
        "--max-epochs",
        "2",
        *extra_args,
        env={"TORCH_DISTRIBUTED_DEBUG": "DETAIL", **os.environ},
    )
    assert ret.success


@pytest.mark.parametrize(
    "accelerators_strategies_devices",
    [
        pytest.param(v, id=runtime_id)
        for v in accelerators_strategies_devices
        if "find_unused_parameters_false" not in (runtime_id := "+".join(map(str, v)))
    ],
)
def test_train_rtd(
    accelerators_strategies_devices: Tuple[str, Optional[str], Optional[int]],
    rtd_model_config: Union[pathlib.Path, str],
    rtd_task_config: pathlib.Path,
    raw_text_path: pathlib.Path,
    script_runner: pytest_console_scripts.ScriptRunner,
    tmp_path: pathlib.Path,
    tokenizer_name_or_path: Union[pathlib.Path, str],
):
    accelerator, strategy, num_devices = accelerators_strategies_devices

    extra_args: List[str] = []
    if strategy is not None:
        extra_args.extend(["--strategy", strategy])
    if num_devices is not None:
        extra_args.extend(["--num-devices", str(num_devices)])
    ret = script_runner.run(
        "zeldarose-transformer",
        "--accelerator",
        accelerator,
        "--config",
        str(rtd_task_config),
        "--tokenizer",
        str(tokenizer_name_or_path),
        "--model-config",
        str(rtd_model_config),
        "--device-batch-size",
        "8",
        "--out-dir",
        str(tmp_path / "train-out"),
        "--cache-dir",
        str(tmp_path / "tokenizer-cache"),
        "--val-text",
        str(raw_text_path),
        str(raw_text_path),
        "--max-epochs",
        "2",
        *extra_args,
        env={"TORCH_DISTRIBUTED_DEBUG": "DETAIL", **os.environ},
    )
    assert ret.success


def test_train_mlm_with_remote_dataset(
    mlm_task_config: pathlib.Path,
    remote_raw_text: str,
    script_runner: pytest_console_scripts.ScriptRunner,
    tmp_path: pathlib.Path,
):
    ret = script_runner.run(
        "zeldarose-transformer",
        "--strategy",
        "ddp_spawn",
        "--num-devices",
        "2",
        "--config",
        str(mlm_task_config),
        "--tokenizer",
        "lgrobol/roberta-minuscule",
        "--model-config",
        "lgrobol/roberta-minuscule",
        "--device-batch-size",
        "8",
        "--out-dir",
        str(tmp_path / "train-out"),
        "--cache-dir",
        str(tmp_path / "tokenizer-cache"),
        "--val-text",
        remote_raw_text,
        remote_raw_text,
        "--max-epochs",
        "2",
    )
    assert ret.success



import click


class ComplexCLI(click.MultiCommand):
    def list_commands(self, ctx):
        return ["tokenizer", "transformer"]

    def get_command(self, ctx, name):
        try:
            mod = __import__(f"zeldarose.train_{name}", None, None, ["cli"])
        except ImportError:
            return
        return mod.main


@click.group(cls=ComplexCLI, help="A straightforward trainer for transformer-based models.")
def cli():
    pass



from abc import ABC, abstractmethod
import pathlib
from typing import Optional, Tuple, Union

import pydantic
import torch
import torchmetrics
import transformers

import pytorch_lightning as pl

from loguru import logger


class TrainConfig(pydantic.BaseModel):
    batch_size: int = 64
    betas: Tuple[float, float] = (0.9, 0.98)
    epsilon: float = 1e-8
    gradient_clipping: Optional[Union[float, int]] = None
    learning_rate: float = 1e-4
    lr_decay_steps: Optional[int] = None
    warmup_steps: int = 0
    weight_decay: Optional[float] = None


class TrainingModule(pl.LightningModule, ABC):
    model: transformers.PreTrainedModel
    training_config: TrainConfig

    @abstractmethod
    def get_data_module(
        self,
        loader_batch_size: int,
        num_workers: int,
        tokenizer: Union[transformers.PreTrainedTokenizer, transformers.PreTrainedTokenizerFast],
        tokenizer_name: str,
        train_path: Union[str, pathlib.Path],
        data_dir: Optional[pathlib.Path] = None,
        val_path: Optional[Union[str, pathlib.Path]] = None,
    ) -> pl.LightningDataModule:
        raise NotImplementedError()

    @abstractmethod
    def save_transformer(
        self,
        save_dir: pathlib.Path,
        tokenizer: Optional[
            Union[transformers.PreTrainedTokenizer, transformers.PreTrainedTokenizerFast]
        ] = None,
    ):
        raise NotImplementedError()
    
    def configure_optimizers(self):
        if self.training_config.weight_decay is not None:
            no_decay = ["bias", "LayerNorm.weight"]
            decay_rate = self.training_config.weight_decay
            optimizer_grouped_parameters = [
                {
                    "params": [
                        p
                        for n, p in self.model.named_parameters()
                        if not any(nd in n for nd in no_decay)
                    ],
                    "weight_decay": decay_rate,
                },
                {
                    "params": [
                        p
                        for n, p in self.model.named_parameters()
                        if any(nd in n for nd in no_decay)
                    ],
                    "weight_decay": 0.0,
                },
            ]
        else:
            decay_rate = 0.0
            optimizer_grouped_parameters = [
                {
                    "params": [p for n, p in self.model.named_parameters()],
                    "weight_decay": 0.0,
                },
            ]

        optimizer = torch.optim.AdamW(
            optimizer_grouped_parameters,
            betas=self.training_config.betas,
            lr=self.training_config.learning_rate,
            eps=self.training_config.epsilon,
            weight_decay=decay_rate,
        )
        if self.training_config.lr_decay_steps:
            if self.training_config.lr_decay_steps == -1:
                num_training_steps = self.trainer.estimated_stepping_batches - self.training_config.warmup_steps
                logger.info(
                    f"Number of lr decay steps set at {num_training_steps} since -1 was asked"
                )
            else:
                if self.training_config.lr_decay_steps > 2 * self.trainer.estimated_stepping_batches:
                    logger.warning(
                        f"Asked for {self.training_config.lr_decay_steps} LR decay steps"
                        f" but the model will only be trained for {self.trainer.estimated_stepping_batches} steps"
                        ", this might be an oversight."
                    )
                num_training_steps = self.training_config.lr_decay_steps

            schedule = transformers.get_linear_schedule_with_warmup(
                optimizer,
                num_warmup_steps=self.training_config.warmup_steps,
                num_training_steps=num_training_steps + self.training_config.warmup_steps,
            )
            schedulers = [{"scheduler": schedule, "interval": "step"}]
        elif self.training_config.warmup_steps > 0:
            schedule = transformers.get_constant_schedule_with_warmup(
                optimizer,
                num_warmup_steps=self.training_config.warmup_steps,
            )
            schedulers = [{"scheduler": schedule, "interval": "step"}]
        else:
            schedulers = []

        return [optimizer], schedulers


class MaskedAccuracy(torchmetrics.Metric):
    full_state_update: bool = False
    higher_is_better: Optional[bool] = True
    is_differentiable = False

    def __init__(self, ignore_index: int = -100, dist_sync_on_step: bool = False):
        super().__init__(compute_on_step=False, dist_sync_on_step=dist_sync_on_step)

        self.ignore_index = ignore_index
        self.correct: torch.Tensor
        self.add_state("correct", default=torch.tensor(0), dist_reduce_fx="sum")
        self.total: torch.Tensor
        self.add_state("total", default=torch.tensor(0), dist_reduce_fx="sum")

    def update(self, preds: torch.Tensor, target: torch.Tensor):  # type: ignore[override]
        assert preds.shape == target.shape
        mask = target.ne(self.ignore_index)
        if mask.any():
            self.correct += preds.eq(target).logical_and(mask).int().sum()
            self.total += mask.sum()

    def compute(self):
        return self.correct.true_divide(self.total)



import pathlib
import typing as ty

import click
import tokenizers
import tokenizers.models
import tokenizers.pre_tokenizers
import tokenizers.processors
import tokenizers.trainers
import transformers

from loguru import logger


@click.command("tokenizer", help="Train a tokenizer.")
@click.argument(
    "raw_texts",
    nargs=-1,
    type=click.Path(resolve_path=True, exists=True, dir_okay=False, path_type=pathlib.Path),
    required=True,
)
@click.option(
    "--max-len",
    type=int,
    default=512,
    help="The maximum number of tokens in a sequence",
    show_default=True,
)
@click.option(
    "--model-name",
    type=str,
    default="muppet",
    help="A name to give to the model",
    show_default=True,
)
@click.option(
    "--out-path",
    default=".",
    type=click.Path(resolve_path=True, file_okay=False, path_type=pathlib.Path),
    help="Where to save the trained model",
)
@click.option(
    "--vocab-size",
    type=int,
    default=4096,
    help="Size of the trained vocabulary",
    show_default=True,
)
def main(
    raw_texts: ty.Collection[pathlib.Path],
    max_len: int,
    model_name: str,
    out_path: pathlib.Path,
    vocab_size: int,
):
    tokenizer = tokenizers.Tokenizer(tokenizers.models.BPE(unk_token="<unk>"))
    tokenizer.pre_tokenizer = tokenizers.pre_tokenizers.ByteLevel()
    # Special tokens hardcoded from RoBERTa's default, see `__init__` in
    # <https://huggingface.co/transformers/_modules/transformers/tokenization_roberta_fast.html#RobertaTokenizerFast>
    # and do not forget to adapt this if we allow other tokenizer configs here
    trainer = tokenizers.trainers.BpeTrainer(
        special_tokens=[
            "<s>",
            "<pad>",
            "</s>",
            "<unk>",
            "<mask>",
        ],
        vocab_size=vocab_size,
    )
    logger.info("Training")
    tokenizer.train(
        [str(t) for t in raw_texts],
        trainer=trainer,
    )
    logger.info("Finalizing")
    tokenizer.post_processor = tokenizers.processors.RobertaProcessing(
        ("</s>", tokenizer.token_to_id("</s>")),
        ("<s>", tokenizer.token_to_id("<s>")),
    )
    tokenizer.enable_truncation(max_length=max_len)
    logger.info("Saving model")
    model_dir = out_path / model_name
    model_dir.mkdir(exist_ok=True, parents=True)
    model_file = model_dir / "tokenizer.json"
    tokenizer.save(str(model_file))
    # TODO: set special tokens instead of hardcoding the default values of RoBERTa (see
    # <https://huggingface.co/transformers/_modules/transformers/tokenization_utils_base.html#SpecialTokensMixin>)
    tranformers_tokenizer = transformers.RobertaTokenizerFast(
        tokenizer_file=str(model_file),
        vocab_file=None,
        merges_file=None,
        max_len=max_len,
    )
    tranformers_tokenizer.save_pretrained(str(model_dir), legacy_format=False)
    # Useless in principle since we don't specify a model here but needed in practice for
    # compatibility with `AutoTokenizer`, see
    # <https://github.com/huggingface/transformers/issues/6368> See also
    # <https://github.com/huggingface/transformers/blob/38f6739cd6c1725ecd75a40d5371483f738097c2/src/transformers/tokenization_utils_base.py#L1670>
    # for some hope that this will be improved some day
    config = transformers.RobertaConfig(
        vocab_size=vocab_size,
        max_position_embeddings=max_len,
        type_vocab_size=1,
    )
    config.to_json_file(str(model_dir / "config.json"))
    logger.info(f"Saved tokenizer in {model_dir}")


if __name__ == "__main__":
    main()



import math
import pathlib
from typing import Any, cast, Dict, NamedTuple, Optional, TYPE_CHECKING, Union

import pydantic
import torch
import torch.jit
import torch.utils.data
import transformers
from loguru import logger
from pytorch_lightning.utilities.rank_zero import rank_zero_only

import zeldarose.datasets.transform
from zeldarose.common import MaskedAccuracy, TrainConfig, TrainingModule


if TYPE_CHECKING:
    import transformers.modeling_outputs


class MaskedTokens(NamedTuple):
    inputs: torch.Tensor
    labels: torch.Tensor


# TODO: How to do whole-word masking?
@torch.jit.script
def mask_tokens(
    inputs: torch.Tensor,
    input_mask_index: int,
    vocabulary_size: int,
    change_ratio: float,
    mask_ratio: float,
    switch_ratio: float,
    keep_mask: Optional[torch.Tensor] = None,
    label_mask_indice: int = -100,
) -> MaskedTokens:
    """Prepare masked tokens inputs/labels for masked language modeling

    Notes
    -----

    - This modifies `inputs` in place, which is not very pure but avoids a (useless in practice)
      copy operation.
    - hf transformers use `-100` for label mask because it's the default ignore index of
      `torch.nn.CrossEntropy`
    """

    labels = inputs.clone()
    # Tells us what to do with each token according to its value `v` in `what_to_do`
    # - `v <= change_ratio` change the token and use it in the loss
    #   - `v <= change_ratio * mask_ratio`: replace with [MASK] ('s id)
    #   - `change_ratio * mask_ratio < v <= change_ratio * (mask_ratio + switch_ratio)`: replace with a random word
    #   - `change_ratio * (mask_ratio + switch_ratio) < v <= change_ratio`: keep as is
    # - `change_ratio < v`: keep as is and don't use in the loss
    # FIXME: layout and device should be inferred here, file an issue
    what_to_do = torch.rand_like(
        labels, layout=torch.strided, dtype=torch.float, device=labels.device
    )
    # This ensures that the internal tokens and the padding is not changed and not used in the loss
    if keep_mask is not None:
        what_to_do.masked_fill_(keep_mask, 1.0)
    preserved_tokens = what_to_do.gt(change_ratio)
    # We only compute loss on masked tokens
    labels.masked_fill_(preserved_tokens, label_mask_indice)

    # replace some input tokens with tokenizer.mask_token ([MASK])
    masked_tokens = what_to_do.le(change_ratio * mask_ratio)
    inputs.masked_fill_(masked_tokens, input_mask_index)

    # Replace masked input tokens with random word
    switched_tokens = what_to_do.le(change_ratio * (mask_ratio + switch_ratio)).logical_and(
        masked_tokens.logical_not()
    )
    random_words = torch.randint_like(labels, vocabulary_size)
    # FIXME: probably still an unnecessary copy here
    inputs[switched_tokens] = random_words[switched_tokens]

    # The rest of the time (10% of the time) we keep the masked input tokens unchanged
    return MaskedTokens(inputs, labels)


class MLMTaskConfig(pydantic.BaseModel):
    change_ratio: float = 0.15
    mask_ratio: float = 0.8
    switch_ratio: float = 0.1


class MLMTrainingModel(TrainingModule):
    def __init__(
        self,
        model: transformers.PreTrainedModel,
        mask_token_index: int,
        vocabulary_size: int,
        training_config: Optional[TrainConfig] = None,
        task_config: Optional[MLMTaskConfig] = None,
    ):
        super().__init__()
        if training_config is not None:
            self.training_config = training_config
        else:
            self.training_config = TrainConfig()
        if task_config is not None:
            self.task_config = task_config
        else:
            self.task_config = MLMTaskConfig()
        logger.info(f"MLM trainer config: {self.training_config}")
        logger.info(f"MLM task config: {self.task_config}")
        self.mask_token_index = mask_token_index
        self.vocabulary_size = vocabulary_size

        self.accuracy = MaskedAccuracy()
        self.model = model

        self.save_hyperparameters("training_config", "task_config")

    def forward(  # type: ignore[override]
        self,
        tokens: torch.Tensor,
        attention_mask: torch.Tensor,
        token_type_ids: torch.Tensor,
        mlm_labels: torch.Tensor,
    ) -> "transformers.modeling_outputs.MaskedLMOutput":
        output = self.model(
            input_ids=tokens,
            attention_mask=attention_mask,
            labels=mlm_labels,
            token_type_ids=token_type_ids,
            return_dict=True,
        )

        return output

    def training_step(  # type: ignore[override]
        self, batch: zeldarose.datasets.transform.TextBatch, batch_idx: int
    ) -> torch.Tensor:
        tokens, attention_mask, internal_tokens_mask, token_type_ids = batch
        with torch.no_grad():
            masked = mask_tokens(
                inputs=tokens,
                change_ratio=self.task_config.change_ratio,
                keep_mask=internal_tokens_mask,
                mask_ratio=self.task_config.mask_ratio,
                input_mask_index=self.mask_token_index,
                switch_ratio=self.task_config.switch_ratio,
                vocabulary_size=self.vocabulary_size,
            )

        outputs = self(
            tokens=masked.inputs,
            attention_mask=attention_mask,
            mlm_labels=masked.labels,
            token_type_ids=token_type_ids,
        )

        loss = outputs.loss

        with torch.no_grad():
            preds = torch.argmax(outputs.logits, dim=-1)
            perplexity = torch.exp(loss)
            self.accuracy(preds, masked.labels)

            self.log(
                "train/loss",
                loss,
                batch_size=tokens.shape[0],
                reduce_fx=torch.mean,
                on_epoch=True,
                sync_dist=True,
            )
            self.log(
                "train/perplexity",
                perplexity,
                batch_size=tokens.shape[0],
                on_epoch=True,
                sync_dist=True,
            )
            self.log(
                "train/accuracy",
                self.accuracy,
                batch_size=tokens.shape[0],
                on_epoch=True,
                on_step=False,
                sync_dist=True,
            )
        return loss

    def validation_step(self, batch: zeldarose.datasets.transform.TextBatch, batch_idx: int):  # type: ignore[override]
        tokens, attention_mask, internal_tokens_mask, token_type_ids = batch
        with torch.no_grad():
            masked = mask_tokens(
                inputs=tokens,
                change_ratio=self.task_config.change_ratio,
                keep_mask=internal_tokens_mask,
                mask_ratio=self.task_config.mask_ratio,
                input_mask_index=self.mask_token_index,
                switch_ratio=self.task_config.switch_ratio,
                vocabulary_size=self.vocabulary_size,
            )

        outputs = self(
            tokens=masked.inputs,
            attention_mask=attention_mask,
            mlm_labels=masked.labels,
            token_type_ids=token_type_ids,
        )

        loss = outputs.loss
        perplexity = torch.exp(loss)

        preds = torch.argmax(outputs.logits, dim=-1)
        self.accuracy(preds, masked.labels)

        self.log("validation/loss", loss, batch_size=tokens.shape[0], sync_dist=True)
        self.log(
            "validation/perplexity",
            perplexity,
            batch_size=tokens.shape[0],
            on_epoch=True,
            sync_dist=True,
        )
        self.log(
            "validation/accuracy",
            self.accuracy,
            batch_size=tokens.shape[0],
            on_epoch=True,
            sync_dist=True,
        )

    def get_data_module(
        self,
        loader_batch_size: int,
        num_workers: int,
        tokenizer: Union[transformers.PreTrainedTokenizer, transformers.PreTrainedTokenizerFast],
        tokenizer_name: str,
        train_path: Union[str, pathlib.Path],
        data_dir: Optional[pathlib.Path] = None,
        val_path: Optional[Union[str, pathlib.Path]] = None,
    ) -> zeldarose.datasets.transform.TextDataModule:
        if (max_length := getattr(self.model.config, "max_position_embeddings")) is None:
            max_length = tokenizer.max_len_single_sentence
        else:
            # FIXME: we shouldn't need num_special_tokens_to_add here
            max_length = min(
                tokenizer.max_len_single_sentence,
                max_length - tokenizer.num_special_tokens_to_add(pair=False),
            )

        return zeldarose.datasets.transform.TextDataModule(
            loader_batch_size=loader_batch_size,
            num_workers=num_workers,
            tokenizer=tokenizer,
            tokenizer_name=tokenizer_name,
            train_path=train_path,
            data_dir=data_dir,
            max_length=max_length,
            val_path=val_path,
        )

    @rank_zero_only
    def save_transformer(
        self,
        save_dir: pathlib.Path,
        tokenizer: Optional[
            Union[transformers.PreTrainedTokenizer, transformers.PreTrainedTokenizerFast]
        ] = None,
    ):
        """Save the wrapped transformer model."""
        save_dir.mkdir(parents=True, exist_ok=True)
        logger.info(f"Saving model to {save_dir}")
        self.model.save_pretrained(str(save_dir))
        if tokenizer is not None:
            logger.info(f"Saving tokenizer to {save_dir}")
            tokenizer.save_pretrained(str(save_dir), legacy_format=not tokenizer.is_fast)


def get_training_model(
    model_config: Optional[Union[str, pathlib.Path]],
    pretrained_model: Optional[Union[str, pathlib.Path]],
    task_config: Optional[Dict[str, Any]],
    tokenizer: Union[transformers.PreTrainedTokenizer, transformers.PreTrainedTokenizerFast],
    training_config: TrainConfig,
) -> MLMTrainingModel:
    if task_config is not None:
        _task_config = MLMTaskConfig.parse_obj(task_config)
    else:
        _task_config = MLMTaskConfig()

    if (
        mask_token_index := cast(Union[int, None], getattr(tokenizer, "mask_token_id", None))
    ) is None:
        mask_token_index = cast(int, tokenizer.convert_tokens_to_ids(tokenizer.mask_token))
    vocabulary_size = tokenizer.vocab_size

    if pretrained_model is not None:
        logger.info(f"Loading pretrained model {pretrained_model!r}")
        model = transformers.AutoModelForMaskedLM.from_pretrained(pretrained_model)
    elif model_config is not None:
        logger.info(f"Loading pretrained config {model_config!r}")
        _model_config = transformers.AutoConfig.from_pretrained(model_config)
        logger.info("Generating model from config")
        # TODO: check the other parameters?
        if vocabulary_size is not None and _model_config.vocab_size != vocabulary_size:
            logger.warning(
                f"Vocabulary size mismatch between model config ({_model_config.vocab_size})"
                f" and pretrained tokenizer ({vocabulary_size}), using {vocabulary_size}."
            )
            _model_config.vocab_size = vocabulary_size
        model = transformers.AutoModelForMaskedLM.from_config(_model_config)
    else:
        raise ValueError("You must provide either a pretrained model or a model config")

    logger.info("Creating MLM training model")

    logger.debug(f"Mask token index: {mask_token_index}")
    training_model = MLMTrainingModel(
        model=model,
        mask_token_index=mask_token_index,
        vocabulary_size=vocabulary_size,
        task_config=_task_config,
        training_config=training_config,
    )

    return training_model



import pathlib
from typing import Any, Collection, List, Mapping, cast, Dict, NamedTuple, Optional, TYPE_CHECKING, Union

import pydantic
import torch
import torch.jit
import torch.utils.data
import transformers
from loguru import logger
from torch.nn.utils.rnn import pad_sequence
from lightning_utilities.core.rank_zero import rank_zero_only
from torchmetrics import SacreBLEUScore

import zeldarose.datasets.mbart
from zeldarose.common import TrainConfig, TrainingModule


if TYPE_CHECKING:
    import transformers.modeling_outputs


class InfilledSent(NamedTuple):
    input_ids: torch.Tensor
    attention_mask: torch.Tensor


# NOTE(2023-02-14): There's a good chance that this is far from optimal and it should be revisited
# but remember that it is far from trivial and in particular, I don't see a way to avoid loops.
def infill_noise(
    change_ratio: float,
    input_ids: torch.Tensor,
    input_mask_id: int,
    poisson_lambda: float,
    padding_id: int = 0,
    keep_mask: Optional[torch.Tensor] = None,
) -> InfilledSent:
    """BART-like span masking with Poisson jumps


    ## Notes

    This is probably somewhat different from the original BART implementation, e.g. here we can have
    several consecutive <MASK> when two masked spans are contiguous. This does not, however,
    significantly change the nature of the task.
    """
    # Vector precomputations for what is easily vectorized
    lengths = input_ids.ne(input_mask_id).sum(dim=-1)
    replace_mask = torch.bernoulli(
        torch.full_like(input_ids, fill_value=change_ratio, dtype=torch.float)
    ).to(torch.bool)
    if keep_mask is not None:
        replace_mask = replace_mask.logical_and(keep_mask.logical_not())
    else:
        keep_mask = torch.zeros_like(input_ids, dtype=torch.bool)
    all_jumps = torch.poisson(
        torch.full_like(input_ids, fill_value=poisson_lambda, dtype=torch.float)
    ).to(torch.long)
    res = []
    # TODO(2023-02-14): This is embarassingly parallel: the sentences can be treated independently
    # BIG LOOP OF HELL
    # You might believe this makes unnecessary copies but it actually Does Not!
    for sent, mask, jump, l, keep in zip(input_ids, replace_mask, all_jumps, lengths, keep_mask):
        current = []
        pos = 0
        while pos < l:
            if mask[pos]:
                current.append(input_mask_id)
                # If it was a 0-length jump, we have to advance to the next token normally
                if jump[pos] == 0:
                    current.append(sent[pos].item())
                    pos += 1
                else:
                    offset = 0
                    # NOTE(2023-02-14): This probably causes our jump distribution to be sub-Poisson (octopus?)
                    while pos + offset < len(sent) and offset < jump[pos]:
                        if keep[pos + offset]:
                            break
                        offset += 1
                    pos += offset
            else:
                current.append(sent[pos].item())
                pos += 1
        res.append(input_ids.new_tensor(current))
    lengths = input_ids.new_tensor([len(t) for t in res])
    # Not optimal, but concise
    attention_mask = (
        torch.arange(lengths.max().item(), device=lengths.device)
        .unsqueeze(0)
        .lt(lengths.unsqueeze(1))
    )
    return InfilledSent(
        input_ids=pad_sequence(res, batch_first=True, padding_value=padding_id),
        attention_mask=attention_mask,
    )


class ForcedBOSTokenLogitsProcessor(transformers.LogitsProcessor):
    r"""
    [`LogitsProcessor`] that enforces the specified token as the first generated token.
    Args:
        bos_token_id (`list[int]`):
            The ids of the tokens to force as the first generated token. Must be as long as the
            batch.
    """

    def __init__(self, bos_token_id: Union[List[int], torch.LongTensor]):
        self.bos_token_id = bos_token_id

    def __call__(self, input_ids: torch.LongTensor, scores: torch.FloatTensor) -> torch.FloatTensor:
        cur_len = input_ids.shape[-1]
        if cur_len == 1:
            num_tokens = scores.shape[1]
            scores[:, :] = -float("inf")
            scores[:, self.bos_token_id] = 0
        return scores


class MBartTaskConfig(pydantic.BaseModel):
    change_ratio: float = 0.3
    denoise_langs: Optional[List[str]]
    denoise_loss_ratio: float = 0.5
    poisson_lambda: float = 3.0
    source_langs: Optional[List[str]]
    target_langs: Optional[List[str]]
    strict_langs: bool = False


class MBartTrainingModel(TrainingModule):
    def __init__(
        self,
        model: transformers.PreTrainedModel,
        mask_token_index: int,
        padding_token_index: int,
        vocabulary_size: int,
        task_config: MBartTaskConfig,
        tokenizer: transformers.PreTrainedTokenizerFast,
        langcode_sub: Optional[Mapping[str,str]] = None,
        training_config: Optional[TrainConfig] = None,
    ):
        super().__init__()
        if training_config is not None:
            self.training_config = training_config
        else:
            self.training_config = TrainConfig()
        self.task_config = task_config
        logger.info(f"mBART trainer config: {self.training_config}")
        logger.info(f"mBART task config: {self.task_config}")
        self.langcode_sub = langcode_sub if langcode_sub is not None else dict()
        self.mask_token_index = mask_token_index
        self.pad_token_index = padding_token_index
        # self.sacrebleu_score = SacreBLEUScore()
        self.vocabulary_size = vocabulary_size

        self.model = model
        self.tokenizer = tokenizer

        self.save_hyperparameters("training_config", "task_config")

    def forward(  # type: ignore[override]
        self,
        attention_mask: torch.Tensor,
        decoder_input_ids: torch.Tensor,
        tokens: torch.Tensor,
        labels: torch.Tensor,
    ) -> "transformers.modeling_outputs.Seq2SeqLMOutput":
        output = self.model(
            attention_mask=attention_mask,
            decoder_input_ids=decoder_input_ids,
            input_ids=tokens,
            labels=labels,
            return_dict=True,
        )

        return output

    def training_step(  # type: ignore[override]
        self, batch: zeldarose.datasets.mbart.TwoMBartBatches, batch_idx: int
    ) -> torch.Tensor:
        denoise, translate = batch

        if denoise is not None:
            with torch.no_grad():
                noisy = infill_noise(
                    change_ratio=self.task_config.change_ratio,
                    input_ids=denoise.input_ids,
                    input_mask_id=self.mask_token_index,
                    keep_mask=denoise.special_tokens_mask,
                    padding_id=self.pad_token_index,
                    poisson_lambda=self.task_config.poisson_lambda,
                )

            denoise_outputs = self(
                tokens=noisy.input_ids,
                decoder_input_ids=denoise.decoder_input_ids,
                labels=denoise.labels,
                attention_mask=noisy.attention_mask,
            )

            denoise_loss = denoise_outputs.loss
            denoise_batch_size = denoise.input_ids.shape[0]

        else:
            denoise_loss = torch.zeros(1, device=self.device)
            denoise_batch_size = 0

        self.log(
            "train/denoise_loss",
            denoise_loss,
            batch_size=denoise_batch_size,
            reduce_fx=torch.mean,
            on_epoch=True,
            sync_dist=True,
        )

        if translate is not None:
            translate_outputs = self(
                tokens=translate.input_ids,
                decoder_input_ids=translate.decoder_input_ids,
                labels=translate.labels,
                attention_mask=translate.attention_mask,
            )

            translate_loss = translate_outputs.loss
            translate_batch_size = translate.input_ids.shape[0]
        else:
            translate_loss = torch.zeros(1, device=self.device)
            translate_batch_size = 0

        self.log(
            "train/translate_loss",
            translate_loss,
            batch_size=translate_batch_size,
            reduce_fx=torch.mean,
            on_epoch=True,
            sync_dist=True,
        )

        translate_coef = translate_batch_size * (1 - self.task_config.denoise_loss_ratio)
        denoise_coef = denoise_batch_size * self.task_config.denoise_loss_ratio
        loss = (translate_coef * translate_loss + denoise_coef * denoise_loss) / (
            translate_coef + denoise_coef
        )

        batch_size = denoise_batch_size + translate_batch_size
        self.log(
            "train/loss",
            loss,
            batch_size=batch_size,
            reduce_fx=torch.mean,
            on_epoch=True,
            sync_dist=True,
        )

        return loss

    def validation_step(self, batch: zeldarose.datasets.mbart.TwoMBartBatches, batch_idx: int):  # type: ignore[override]
        denoise, translate = batch

        if denoise is not None:
            noisy = infill_noise(
                change_ratio=self.task_config.change_ratio,
                input_ids=denoise.input_ids,
                input_mask_id=self.mask_token_index,
                keep_mask=denoise.special_tokens_mask,
                padding_id=self.pad_token_index,
                poisson_lambda=self.task_config.poisson_lambda,
            )

            denoise_outputs = self(
                tokens=noisy.input_ids,
                decoder_input_ids=denoise.decoder_input_ids,
                labels=denoise.labels,
                attention_mask=noisy.attention_mask,
            )

            denoise_loss = denoise_outputs.loss
            denoise_batch_size = denoise.input_ids.shape[0]

        else:
            denoise_loss = torch.zeros(1, device=self.device)
            denoise_batch_size = 0

        self.log(
            "validation/denoise_loss",
            denoise_loss,
            batch_size=denoise_batch_size,
            reduce_fx=torch.mean,
            on_epoch=True,
            sync_dist=True,
        )

        if translate is not None:
            translate_outputs = self(
                tokens=translate.input_ids,
                decoder_input_ids=translate.decoder_input_ids,
                labels=translate.labels,
                attention_mask=translate.attention_mask,
            )

            translate_loss = translate_outputs.loss
            translate_batch_size = translate.input_ids.shape[0]

            # generated_ids = self.model.generate(
            #     input_ids=translate.input_ids,
            #     logits_processor=transformers.LogitsProcessorList(
            #         [
            #             ForcedBOSTokenLogitsProcessor(
            #                 cast(torch.LongTensor, translate.decoder_input_ids[:, 1])
            #             )
            #         ]
            #     ),
            # )
            # generated_txt = self.tokenizer.batch_decode(generated_ids, skip_special_tokens=True)
            # self.sacrebleu_score(generated_txt, translate.tgt_text)

        else:
            translate_loss = torch.zeros(1, device=self.device)
            translate_batch_size = 0

        self.log(
            "validation/translate_loss",
            denoise_loss,
            batch_size=translate_batch_size,
            reduce_fx=torch.mean,
            on_epoch=True,
            sync_dist=True,
        )
        # self.log("validation/sacrebleu", self.sacrebleu_score, on_epoch=True, on_step=False)

        translate_coef = translate_batch_size * (1 - self.task_config.denoise_loss_ratio)
        denoise_coef = denoise_batch_size * self.task_config.denoise_loss_ratio
        loss = (translate_coef * translate_loss + denoise_coef * denoise_loss) / (
            translate_coef + denoise_coef
        )

        batch_size = denoise_batch_size + translate_batch_size
        self.log(
            "validation/loss",
            loss,
            batch_size=batch_size,
            reduce_fx=torch.mean,
            on_epoch=True,
            sync_dist=True,
        )

    def get_data_module(
        self,
        loader_batch_size: int,
        num_workers: int,
        tokenizer: Union[transformers.PreTrainedTokenizer, transformers.PreTrainedTokenizerFast],
        tokenizer_name: str,
        train_path: Union[str, pathlib.Path],
        data_dir: Optional[pathlib.Path] = None,
        val_path: Optional[Union[str, pathlib.Path]] = None,
    ) -> zeldarose.datasets.mbart.MBartDataModule:
        if (max_length := getattr(self.model.config, "max_position_embeddings")) is None:
            max_length = tokenizer.max_len_single_sentence
        else:
            # FIXME: we shouldn't need num_special_tokens_to_add here
            max_length = min(
                tokenizer.max_len_single_sentence,
                max_length - tokenizer.num_special_tokens_to_add(pair=False),
            )

        return zeldarose.datasets.mbart.MBartDataModule(
            data_dir=data_dir,
            denoise_langs=(
                self.task_config.denoise_langs if self.task_config.denoise_langs is not None else []
            ),
            langcode_sub=self.langcode_sub,
            loader_batch_size=loader_batch_size,
            max_length=max_length,
            num_workers=num_workers,
            source_langs=(
                self.task_config.source_langs if self.task_config.source_langs is not None else []
            ),
            tokenizer=tokenizer,
            tokenizer_name=tokenizer_name,
            train_path=train_path,
            target_langs=(
                self.task_config.target_langs if self.task_config.target_langs is not None else []
            ),
            val_path=val_path,
        )

    @rank_zero_only
    def save_transformer(
        self,
        save_dir: pathlib.Path,
        tokenizer: Optional[
            Union[transformers.PreTrainedTokenizer, transformers.PreTrainedTokenizerFast]
        ] = None,
    ):
        """Save the wrapped transformer model."""
        save_dir.mkdir(parents=True, exist_ok=True)
        logger.info(f"Saving model to {save_dir}")
        self.model.save_pretrained(str(save_dir))
        if tokenizer is not None:
            logger.info(f"Saving tokenizer to {save_dir}")
            tokenizer.save_pretrained(str(save_dir), legacy_format=not tokenizer.is_fast)


def get_training_model(
    model_config: Optional[Union[str, pathlib.Path]],
    pretrained_model: Optional[Union[str, pathlib.Path]],
    task_config: Dict[str, Any],
    tokenizer: transformers.PreTrainedTokenizerFast,
    training_config: TrainConfig,
) -> MBartTrainingModel:
    if not hasattr(tokenizer, "lang_code_to_id"):
        raise ValueError(
            "The tokenizer for mBART training must be multilingual and have a `lang_code_to_id` attribute"
        )

    _task_config = MBartTaskConfig.parse_obj(task_config)

    if (
        mask_token_index := cast(Union[int, None], getattr(tokenizer, "mask_token_id", None))
    ) is None:
        mask_token_index = cast(int, tokenizer.convert_tokens_to_ids(tokenizer.mask_token))
    vocabulary_size = tokenizer.vocab_size

    if pretrained_model is not None:
        logger.info(f"Loading pretrained model {pretrained_model!r}")
        model = transformers.AutoModelForSeq2SeqLM.from_pretrained(pretrained_model)
    elif model_config is not None:
        logger.info(f"Loading pretrained config {model_config!r}")
        _model_config = transformers.AutoConfig.from_pretrained(model_config)
        logger.info("Generating model from config")
        # TODO: check the other parameters?
        if vocabulary_size is not None and _model_config.vocab_size != vocabulary_size:
            logger.warning(
                f"Vocabulary size mismatch between model config ({_model_config.vocab_size})"
                f" and pretrained tokenizer ({vocabulary_size}), using {vocabulary_size}."
            )
            _model_config.vocab_size = vocabulary_size
        model = transformers.AutoModelForSeq2SeqLM.from_config(_model_config)
    else:
        raise ValueError("You must provide either a pretrained model or a model config")

    if tokenizer.mask_token is None:
        logger.info("Adding a mask token to the model that's missing one.")
        tokenizer.add_special_tokens({"mask_token": "<mask>"})
        model.resize_token_embeddings(len(tokenizer))

    langcode_sub: Dict[str, str] = dict()
    if not _task_config.strict_langs:
        logger.debug("Checking match between task and tokenizer langs")
        all_langs = set().union(
            *(
                l
                for l in (
                    _task_config.denoise_langs,
                    _task_config.source_langs,
                    _task_config.target_langs,
                )
                if l is not None
            )
        )
        for lang in all_langs:
            # FIXME: we need ignores here because (at least) Pyright doesn't take the hasattr into account
            if (substitute_lang := match_lang(lang, tokenizer.lang_code_to_id)) is None:  # type: ignore
                logger.warning(
                    f"Language {lang} is unknown of the tokenizer, adding it and resizing the model vocabulary."
                )
                tokenizer.add_tokens(lang, special_tokens=True)
                lang_id = cast(int, tokenizer.convert_tokens_to_ids(lang))
                tokenizer.lang_code_to_id[lang] = lang_id  # type: ignore
                model.resize_token_embeddings(len(tokenizer))
            else:
                if substitute_lang != lang:
                    logger.info(f"Adding {lang} as an alias to {substitute_lang} in the tokenizer.")
                tokenizer.lang_code_to_id[lang] = tokenizer.lang_code_to_id[substitute_lang]  # type: ignore
                langcode_sub[lang] = substitute_lang

    logger.info("Creating mBART training model")

    logger.debug(f"Mask token index: {mask_token_index}")
    training_model = MBartTrainingModel(
        langcode_sub=langcode_sub,
        model=model,
        mask_token_index=mask_token_index,
        padding_token_index=cast(int, tokenizer.pad_token_id),
        task_config=_task_config,
        tokenizer=tokenizer,
        training_config=training_config,
        vocabulary_size=vocabulary_size,
    )

    return training_model


def match_lang(lang: str, available: Collection[str]) -> Optional[str]:
    if lang in available:
        return lang
    logger.debug(f"{lang} not found in tokenizer langs: {available}. Looking for a substitute.")
    substitutes = [
        model_lang
        for model_lang in available
        if model_lang.split("_", maxsplit=1)[0] == lang
        or lang.split("_", maxsplit=1)[0] == model_lang
    ]
    if len(substitutes) > 1:
        raise ValueError(f"Multiple tokenizer langs would fit {lang}: {substitutes}")
    elif len(substitutes) == 0:
        return None
    else:
        return substitutes[0]



import pathlib
from typing import Any, cast, Dict, NamedTuple, List, Literal, Optional, TYPE_CHECKING, Union

import pydantic
import pytorch_lightning as pl
import torch
import torch.jit
import torch.utils.data
import transformers
from loguru import logger
from lightning_utilities.core.rank_zero import rank_zero_only

import zeldarose.datasets.transform
from zeldarose.common import MaskedAccuracy, TrainConfig, TrainingModule
from zeldarose.utils import (
    OneWayShareTransformersEmbeddingsCallback,
    ShareTransformersEmbeddingsCallback,
)


if TYPE_CHECKING:
    import transformers.modeling_outputs


class MaskedTokens(NamedTuple):
    inputs: torch.Tensor
    labels: torch.Tensor


class RTDOutput(NamedTuple):
    discriminator_output: transformers.modeling_outputs.TokenClassifierOutput
    generator_predictions: torch.Tensor
    discriminator_predictions: torch.Tensor
    generator_output: transformers.modeling_outputs.MaskedLMOutput
    rtd_labels: torch.Tensor


@torch.jit.script
def mask_tokens(
    tokens: torch.Tensor,
    input_mask_index: int,
    mask_ratio: float,
    keep_mask: Optional[torch.Tensor] = None,
    label_mask_index: int = -100,
) -> MaskedTokens:
    """Prepare masked tokens inputs/labels for masked language modeling

    ## Notes

    - The default for label mask is `-100` because it's the default ignore index of
      `torch.nn.CrossEntropy`
    """

    inputs = tokens.clone()
    labels = tokens.clone()
    # Tells us what to do with each token according to its value `v` in `what_to_do`
    # - `v <= mask_ratio` mask the token and use it in the loss
    # - `mask_ratio < v`: keep as is and don't use in the loss
    # FIXME: layout and device should be inferred here, file an issue
    what_to_do = torch.rand_like(
        inputs, layout=torch.strided, dtype=torch.float, device=labels.device
    )
    # This ensures that the internal tokens and the padding is not changed and not used in the loss
    if keep_mask is not None:
        what_to_do.masked_fill_(keep_mask, 1.0)
    preserved_tokens = what_to_do.gt(mask_ratio)
    # We only compute loss on masked tokens
    labels.masked_fill_(preserved_tokens, label_mask_index)

    # replace some input tokens with tokenizer.mask_token ([MASK])
    masked_tokens = what_to_do.le(mask_ratio)
    inputs.masked_fill_(masked_tokens, input_mask_index)

    return MaskedTokens(inputs, labels)


class RTDTaskConfig(pydantic.BaseModel):
    discriminator_loss_weight: float = 1.0
    embeddings_sharing: Optional[Literal["deberta", "electra"]] = None
    mask_ratio: float = 0.15


class RTDTrainingModel(TrainingModule):
    def __init__(
        self,
        discriminator: transformers.PreTrainedModel,
        generator: transformers.PreTrainedModel,
        mask_token_index: int,
        vocabulary_size: int,
        training_config: Optional[TrainConfig] = None,
        task_config: Optional[RTDTaskConfig] = None,
    ):
        super().__init__()
        if training_config is not None:
            self.training_config = training_config
        else:
            self.training_config = TrainConfig()
        if task_config is not None:
            self.task_config = task_config
        else:
            self.task_config = RTDTaskConfig()
        logger.info(f"RTD trainer config: {self.training_config}")
        logger.info(f"RTD task config: {self.task_config}")
        self.mask_token_index = mask_token_index
        self.vocabulary_size = vocabulary_size

        self.generator_accuracy = MaskedAccuracy()
        self.discriminator_accuracy = MaskedAccuracy()
        self.generator = generator
        self.discriminator = discriminator

        self.save_hyperparameters("training_config", "task_config")

    def forward(  # type: ignore[override]
        self,
        attention_mask: torch.Tensor,
        mlm_labels: torch.Tensor,
        internal_tokens_mask: torch.Tensor,
        tokens: torch.Tensor,
        token_type_ids: torch.Tensor,
    ) -> RTDOutput:
        generator_output = cast(
            transformers.modeling_outputs.MaskedLMOutput,
            self.generator(
                input_ids=tokens,
                attention_mask=attention_mask,
                labels=mlm_labels,
                token_type_ids=token_type_ids,
                return_dict=True,
            ),
        )

        generator_predictions = torch.argmax(generator_output.logits, dim=-1).detach()
        rtd_labels = generator_predictions.ne(tokens).to(torch.long)
        rtd_labels.masked_fill(internal_tokens_mask, -100)
        rtd_labels.masked_fill(attention_mask, -100)

        discriminator_output = cast(
            transformers.modeling_outputs.TokenClassifierOutput,
            self.discriminator(
                input_ids=tokens,
                attention_mask=attention_mask,
                labels=rtd_labels,
                token_type_ids=token_type_ids,
                return_dict=True,
            ),
        )
        discriminator_predictions = discriminator_output.logits.argmax(dim=-1)

        return RTDOutput(
            discriminator_output=discriminator_output,
            discriminator_predictions=discriminator_predictions,
            generator_output=generator_output,
            generator_predictions=generator_predictions,
            rtd_labels=rtd_labels,
        )

    def training_step(self, batch: zeldarose.datasets.transform.TextBatch, batch_idx: int) -> torch.Tensor:  # type: ignore[override]
        tokens, attention_mask, internal_tokens_mask, token_type_ids = batch
        with torch.no_grad():
            masked = mask_tokens(
                tokens=tokens,
                keep_mask=internal_tokens_mask,
                label_mask_index=-100,
                mask_ratio=self.task_config.mask_ratio,
                input_mask_index=self.mask_token_index,
            )

        outputs: RTDOutput = self(
            tokens=masked.inputs,
            attention_mask=attention_mask,
            mlm_labels=masked.labels,
            internal_tokens_mask=internal_tokens_mask,
            token_type_ids=token_type_ids,
        )

        combined_loss = cast(
            torch.Tensor, outputs.generator_output.loss
        ) + self.task_config.discriminator_loss_weight * cast(
            torch.Tensor, outputs.discriminator_output.loss
        )

        with torch.no_grad():
            generator_perplexity = torch.exp(cast(torch.Tensor, outputs.generator_output.loss))
            self.generator_accuracy(outputs.generator_predictions, masked.labels)

            self.log(
                "train/generator_loss",
                cast(torch.Tensor, outputs.generator_output.loss),
                reduce_fx=torch.mean,
                on_epoch=True,
                sync_dist=True,
            )
            self.log(
                "train/generator_perplexity",
                generator_perplexity,
                on_epoch=True,
                sync_dist=True,
            )
            self.log(
                "train/generator_accuracy",
                self.generator_accuracy,
                on_epoch=True,
                on_step=False,
                sync_dist=True,
            )

            self.discriminator_accuracy(outputs.discriminator_predictions, outputs.rtd_labels)
            self.log(
                "train/discriminator_loss",
                cast(torch.Tensor, outputs.discriminator_output.loss),
                reduce_fx=torch.mean,
                on_epoch=True,
                sync_dist=True,
            )
            self.log(
                "train/discriminator_accuracy",
                self.discriminator_accuracy,
                on_epoch=True,
                on_step=False,
                sync_dist=True,
            )

            self.log(
                "train/combined_loss",
                combined_loss,
                reduce_fx=torch.mean,
                on_epoch=True,
                sync_dist=True,
            )

        return combined_loss

    def validation_step(self, batch: zeldarose.datasets.transform.TextBatch, batch_idx: int):  # type: ignore[override]
        tokens, attention_mask, internal_tokens_mask, token_type_ids = batch
        with torch.no_grad():
            masked = mask_tokens(
                tokens=tokens,
                keep_mask=internal_tokens_mask,
                label_mask_index=-100,
                mask_ratio=self.task_config.mask_ratio,
                input_mask_index=self.mask_token_index,
            )

        outputs: RTDOutput = self(
            tokens=masked.inputs,
            attention_mask=attention_mask,
            mlm_labels=masked.labels,
            internal_tokens_mask=internal_tokens_mask,
            token_type_ids=token_type_ids,
        )
        generator_perplexity = torch.exp(cast(torch.Tensor, outputs.generator_output.loss))
        self.generator_accuracy(outputs.generator_predictions, masked.labels)

        self.log(
            "validation/generator_loss",
            cast(torch.Tensor, outputs.generator_output.loss),
            reduce_fx=torch.mean,
            on_epoch=True,
            sync_dist=True,
        )
        self.log(
            "validation/generator_perplexity",
            generator_perplexity,
            on_epoch=True,
            sync_dist=True,
        )
        self.log(
            "validation/generator_accuracy",
            self.generator_accuracy,
            on_epoch=True,
            sync_dist=True,
        )

        self.discriminator_accuracy(outputs.discriminator_predictions, outputs.rtd_labels)
        self.log(
            "validation/discriminator_loss",
            cast(torch.Tensor, outputs.discriminator_output.loss),
            reduce_fx=torch.mean,
            on_epoch=True,
            sync_dist=True,
        )
        self.log(
            "validation/discriminator_accuracy",
            self.discriminator_accuracy,
            on_epoch=True,
            sync_dist=True,
        )

    def configure_callbacks(self):
        callbacks: List[pl.Callback] = []
        if self.task_config.embeddings_sharing == "electra":
            callbacks.append(
                ShareTransformersEmbeddingsCallback(
                    leader=self.generator, follower=self.discriminator
                )
            )
        elif self.task_config.embeddings_sharing == "deberta":
            callbacks.append(
                OneWayShareTransformersEmbeddingsCallback(
                    leader=self.generator, follower=self.discriminator
                )
            )
        return callbacks

    def configure_optimizers(self):
        named_parameters = [
            (n, p)
            for model in [self.generator, self.discriminator]
            for n, p in model.named_parameters()
        ]
        if self.training_config.weight_decay is not None:
            no_decay = ["bias", "LayerNorm.weight"]
            decay_rate = self.training_config.weight_decay
            optimizer_grouped_parameters = [
                {
                    "params": [
                        p for n, p in named_parameters if not any(nd in n for nd in no_decay)
                    ],
                    "weight_decay": decay_rate,
                },
                {
                    "params": [p for n, p in named_parameters if any(nd in n for nd in no_decay)],
                    "weight_decay": 0.0,
                },
            ]
        else:
            decay_rate = 0.0
            optimizer_grouped_parameters = [
                {
                    "params": [p for _, p in named_parameters],
                    "weight_decay": 0.0,
                },
            ]

        optimizer = torch.optim.AdamW(
            optimizer_grouped_parameters,
            betas=self.training_config.betas,
            lr=self.training_config.learning_rate,
            eps=self.training_config.epsilon,
            weight_decay=decay_rate,
        )
        if self.training_config.lr_decay_steps:
            if self.training_config.lr_decay_steps > 2 * self.trainer.max_steps:
                logger.warning(
                    f"Asked for {self.training_config.lr_decay_steps} LR decay steps"
                    f" but the model will only be trained for {self.trainer.max_steps} steps"
                    ", this might be an oversight."
                )
            if self.training_config.lr_decay_steps == -1:
                num_training_steps = self.trainer.max_steps - self.training_config.warmup_steps
                logger.info(
                    f"Number of lr decay steps set at {num_training_steps} since -1 was asked"
                )
            else:
                num_training_steps = self.training_config.lr_decay_steps

            schedule = transformers.get_linear_schedule_with_warmup(
                optimizer,
                num_warmup_steps=self.training_config.warmup_steps,
                num_training_steps=num_training_steps + self.training_config.warmup_steps,
            )
            schedulers = [{"scheduler": schedule, "interval": "step"}]
        elif self.training_config.warmup_steps > 0:
            schedule = transformers.get_constant_schedule_with_warmup(
                optimizer,
                num_warmup_steps=self.training_config.warmup_steps,
            )
            schedulers = [{"scheduler": schedule, "interval": "step"}]
        else:
            schedulers = []

        return [optimizer], schedulers

    def get_data_module(
        self,
        loader_batch_size: int,
        num_workers: int,
        tokenizer: Union[transformers.PreTrainedTokenizer, transformers.PreTrainedTokenizerFast],
        tokenizer_name: str,
        train_path: Union[str, pathlib.Path],
        data_dir: Optional[pathlib.Path] = None,
        val_path: Optional[Union[str, pathlib.Path]] = None,
    ) -> zeldarose.datasets.transform.TextDataModule:
        max_length = min(
            getattr(self.generator.config, "max_position_embeddings", float("inf")),
            getattr(self.discriminator.config, "max_position_embeddings", float("inf")),
        )

        if max_length == float("inf"):
            max_length = tokenizer.max_len_single_sentence
        else:
            # FIXME: we shouldn't need num_special_tokens_to_add here
            max_length = min(
                tokenizer.max_len_single_sentence,
                cast(int, max_length) - tokenizer.num_special_tokens_to_add(pair=False),
            )

        return zeldarose.datasets.transform.TextDataModule(
            loader_batch_size=loader_batch_size,
            num_workers=num_workers,
            tokenizer=tokenizer,
            tokenizer_name=tokenizer_name,
            train_path=train_path,
            data_dir=data_dir,
            max_length=max_length,
            val_path=val_path,
        )

    @rank_zero_only
    def save_transformer(
        self,
        save_dir: pathlib.Path,
        tokenizer: Optional[transformers.PreTrainedTokenizer] = None,
    ):
        """Save the wrapped transformer model."""
        save_dir.mkdir(parents=True, exist_ok=True)
        logger.info(f"Saving models to {save_dir}")
        for model_name, model in [
            ("discriminator", self.discriminator),
            ("generator", self.generator),
        ]:
            save_subdir = save_dir / model_name
            model.save_pretrained(str(save_subdir))
            if tokenizer is not None:
                logger.info(f"Saving tokenizer to {save_subdir}")
                tokenizer.save_pretrained(str(save_subdir), legacy_format=not tokenizer.is_fast)


def get_training_model(
    model_config: Optional[str],
    pretrained_model: Optional[str],
    task_config: Optional[Dict[str, Any]],
    tokenizer: Union[transformers.PreTrainedTokenizer, transformers.PreTrainedTokenizerFast],
    training_config: TrainConfig,
) -> RTDTrainingModel:
    if task_config is not None:
        _task_config = RTDTaskConfig.parse_obj(task_config)
    else:
        _task_config = RTDTaskConfig()

    if (
        mask_token_index := cast(Union[int, None], getattr(tokenizer, "mask_token_id", None))
    ) is None:
        mask_token_index = cast(int, tokenizer.convert_tokens_to_ids(tokenizer.mask_token))
    vocabulary_size = tokenizer.vocab_size

    if pretrained_model is not None:
        pretrained_discriminator, pretrained_generator = pretrained_model.split(",")
        logger.info(f"Loading pretrained discriminator {pretrained_discriminator!r}")
        discriminator = transformers.AutoModelForTokenClassification.from_pretrained(
            pretrained_discriminator
        )
        logger.info(f"Loading pretrained generator {pretrained_generator!r}")
        generator = transformers.AutoModelForMaskedLM.from_pretrained(pretrained_generator)
    elif model_config is not None:
        (
            discriminator_config_path,
            generator_config_path,
        ) = model_config.split(",")
        logger.info(f"Loading discriminator config {discriminator_config_path!r}")
        discriminator_config = transformers.AutoConfig.from_pretrained(discriminator_config_path)
        if vocabulary_size is not None and discriminator_config.vocab_size != vocabulary_size:
            logger.warning(
                f"Vocabulary size mismatch between discriminator config ({discriminator_config.vocab_size})"
                f" and pretrained tokenizer ({vocabulary_size}), using {vocabulary_size}."
            )
            discriminator_config.vocab_size = vocabulary_size
        logger.info(f"Loading generator config {generator_config_path,!r}")
        generator_config = transformers.AutoConfig.from_pretrained(generator_config_path)
        if vocabulary_size is not None and generator_config.vocab_size != vocabulary_size:
            logger.warning(
                f"Vocabulary size mismatch between generator config ({generator_config.vocab_size})"
                f" and pretrained tokenizer ({vocabulary_size}), using {vocabulary_size}."
            )
            generator_config.vocab_size = vocabulary_size
        logger.info("Generating discriminator from config")
        discriminator = transformers.AutoModelForMaskedLM.from_config(discriminator_config)
        logger.info("Generating generator from config")
        generator = transformers.AutoModelForMaskedLM.from_config(generator_config)
    else:
        raise ValueError("You must provide either pretrained models or model configs")
    # 2 is the default
    # (<https://github.com/huggingface/transformers/blob/e68c3756fea7c811d02b8470539ae17ec3ec0e71/src/transformers/configuration_utils.py#L302>)
    # but it could have been overriden
    if not discriminator.config.num_labels == 2:
        raise ValueError(
            f"RTD discriminator must have exactly 2 tokens classes, found {discriminator.config.num_labels}"
        )
    if discriminator.config.vocab_size != generator.config.vocab_size:
        raise ValueError(
            "Vocabulary size mismatch between discriminator and generator:"
            f" {discriminator.config.vocab_size} vs {generator.config.vocab_size}"
        )

    logger.info("Creating RTD training model")

    logger.debug(f"Mask token index: {mask_token_index}")
    training_model = RTDTrainingModel(
        discriminator=discriminator,
        generator=generator,
        mask_token_index=mask_token_index,
        vocabulary_size=vocabulary_size,
        task_config=_task_config,
        training_config=training_config,
    )

    return training_model



import logging
import math
import os
import pathlib
from types import ModuleType
import warnings

from typing import Any, Dict, List, Literal, Optional, Union, cast

import click
import pytorch_lightning as pl
from pytorch_lightning import (
    callbacks as pl_callbacks,
    profilers as pl_profilers,
    strategies as pl_strategies,
)
from pytorch_lightning.utilities import types as pl_types
import rich
import tomli
import torch
import transformers

from lightning_utilities.core.rank_zero import rank_zero_only
from loguru import logger
from zeldarose.common import TrainConfig, TrainingModule
from zeldarose.tasks import mbart, mlm, rtd


def setup_logging(
    console: rich.console.Console,
    verbose: bool,
    log_file: Optional[pathlib.Path] = None,
    replace_warnings: bool = True,
):
    logger.remove(0)  # Remove the default logger
    if "SLURM_JOB_ID" in os.environ:
        local_id = os.environ.get("SLURM_LOCALID", "someproc")
        node_name = os.environ.get("SLURMD_NODENAME", "somenode")
        appname = (
            f"zeldarose ({os.environ.get('SLURM_PROCID', 'somerank')} [{local_id}@{node_name}])"
        )
    else:
        appname = "zeldarose"

    if verbose:
        log_level = "DEBUG"
        log_fmt = (
            f"\\[{appname}]"
            " [green]{time:YYYY-MM-DD HH:mm:ss.SSS}[/green] | [blue]{level: <8}[/blue] |"
            " {message}"
        )
    else:
        logging.getLogger(None).setLevel(logging.CRITICAL)
        log_level = "INFO"
        log_fmt = (
            f"\\[{appname}]"
            " [green]{time:YYYY-MM-DD}T{time:HH:mm:ss}[/green] {level} "
            " {message}"
        )

    logger.add(
        lambda m: console.print(m, end=""),
        colorize=True,
        format=log_fmt,
        level=log_level,
    )

    if log_file:
        logger.add(
            log_file,
            colorize=False,
            format=(f"[{appname}] {{time:YYYY-MM-DD HH:mm:ss.SSS}} | {{level: <8}} | {{message}}"),
            level="DEBUG",
        )

    # Deal with stdlib.logging

    class InterceptHandler(logging.Handler):
        def __init__(self, wrapped_name: Optional[str] = None, *args, **kwargs):
            self.wrapped_name = wrapped_name
            super().__init__(*args, **kwargs)

        def emit(self, record):
            # Get corresponding Loguru level if it exists
            try:
                level = logger.level(record.levelname).name
            except ValueError:
                level = record.levelno

            # Find caller from where originated the logged message
            frame, depth = logging.currentframe(), 2
            while frame is not None and frame.f_code.co_filename == logging.__file__:
                frame = frame.f_back
                depth += 1

            if frame is None:
                warnings.warn(
                    "Catching calls to logging is impossible in stackless environment, logging from external libraries might be lost."
                )
            else:
                if self.wrapped_name is not None:
                    logger.opt(depth=depth, exception=record.exc_info).log(
                        level, f"[bold]{self.wrapped_name} says:[/bold] {record.getMessage()}"
                    )
                else:
                    logger.opt(depth=depth, exception=record.exc_info).log(
                        level, record.getMessage()
                    )

    for libname in (
        "datasets",
        "huggingface_hub",
        "lightning_fabric",
        "pytorch_lightning",
        "torch",
        "torchmetrics",
        "transformers",
    ):
        lib_logger = logging.getLogger(libname)
        # FIXME: ugly, but is there a better way? What if they rely on having other handlers?
        if lib_logger.handlers:
            lib_logger.handlers = []
        lib_logger.addHandler(InterceptHandler(libname))
        logger.info(f"Intercepting logging from {libname}")

    # Deal with stdlib.warnings
    def showwarning(message, category, filename, lineno, file=None, line=None):
        logger.warning(warnings.formatwarning(message, category, filename, lineno, None).strip())

    if replace_warnings:
        warnings.showwarning = showwarning


class SavePretrainedModelCallback(pl.Callback):
    def __init__(
        self,
        save_dir: pathlib.Path,
        tokenizer: Union[transformers.PreTrainedTokenizer, transformers.PreTrainedTokenizerFast],
        epoch_period: Optional[int] = 1,
        step_period: Optional[int] = None,
    ):
        self.epoch_period = epoch_period
        self.step_period = step_period
        self.save_dir = save_dir
        self.tokenizer = tokenizer

    @rank_zero_only
    def on_train_batch_end(
        self,
        trainer: pl.Trainer,
        pl_module: TrainingModule,
        outputs: pl_types.STEP_OUTPUT,
        batch: Any,
        batch_idx: int,
    ):
        step = trainer.global_step
        if self.step_period is not None and step % self.step_period == 0:
            step_save_dir = self.save_dir / f"step_{step}"
            logger.info(f"Saving intermediate model to {step_save_dir}")
            pl_module.save_transformer(step_save_dir, self.tokenizer)

    @rank_zero_only
    def on_train_epoch_end(self, trainer: pl.Trainer, pl_module: TrainingModule):
        epoch = trainer.current_epoch
        if self.epoch_period is not None and epoch % self.epoch_period == 0:
            epoch_save_dir = self.save_dir / f"epoch_{trainer.current_epoch}"
            logger.info(f"Saving intermediate model to {epoch_save_dir}")
            pl_module.save_transformer(epoch_save_dir, self.tokenizer)


# TODO: allow reading all these from a config file (except perhaps for paths)
@click.command("transformer", help="Train a transformer model.")
@click.argument("train_data")
@click.option(
    "--accelerator",
    default="auto",
    show_default=True,
    type=str,
    help="The lightning accelerator to use (see lightning doc).",
)
@click.option(
    "--cache-dir",
    type=click.Path(resolve_path=True, file_okay=False, path_type=pathlib.Path),
    help="Where to cache the input data",
)
@click.option(
    "--checkpoint",
    "checkpoint",
    type=click.Path(resolve_path=True, dir_okay=False, exists=True, path_type=pathlib.Path),
    help="A checkpoint to restore the training state from",
)
@click.option(
    "--config",
    "config_path",
    type=click.Path(resolve_path=True, dir_okay=False, exists=True, path_type=pathlib.Path),
    help="A config file (in TOML format)",
)
@click.option(
    "--device-batch-size",
    type=click.IntRange(0),
    help=(
        "Number of samples in a processing batch"
        " (must be a divisor of training bath size, defaults to training batch size)"
    ),
)
@click.option(
    "--guess-batch-size",
    is_flag=True,
    help=(
        "Try to find the max device batch size automatically"
        " (ignored if --device-batch-size is provided)"
    ),
)
@click.option(
    "--max-epochs",
    type=click.IntRange(0),
    help="How many epochs to train for",
)
@click.option(
    "--max-steps",
    type=click.IntRange(0),
    help="How many steps to train for",
)
@click.option(
    "--model-config",
    "model_config_path",
    type=str,
    default="roberta-base",
    show_default=True,
    metavar="NAME_OR_PATH",
    help="A config name or path to create a muppet from scratch",
)
@click.option(
    "--model-name",
    type=str,
    default="muppet",
    metavar="NAME",
    help="A name to give to the model",
)
@click.option(
    "--num-devices",
    default=os.environ.get("SLURM_NTASKS", 1),
    type=click.IntRange(1),
    help="How many devices to train on. If `accelerator` is `cpu`, this is the number of processes",
)
@click.option(
    "--num-nodes",
    type=click.IntRange(0),
    default=os.environ.get("SLURM_JOB_NUM_NODES", 1),
    help="How many nodes to train on (for clusters), defaults to $SLURM_JOB_NUM_NODES if on SLURM and 1 otherwise",
)
@click.option(
    "--num-workers",
    type=click.IntRange(0),
    default=0,
    help="How many data loading workers to use",
)
@click.option(
    "--out-dir",
    default=".",
    type=click.Path(resolve_path=True, file_okay=False, path_type=pathlib.Path),
    help="Where to save the trained model. (defaults to cwd)",
)
@click.option(
    "--precision",
    type=click.Choice(["64", "32", "16", "bf16"]),
    help="The precision of float for traingin",
)
@click.option(
    "--pretrained-model",
    type=str,
    help="A pretrained model to fine-tune",
    metavar="NAME_OR_PATH",
)
@click.option(
    "--save-period",
    "--epoch-save-period",
    "epoch_save_period",
    type=click.IntRange(0),
    help="The number of epoch between intermediate model saving",
)
@click.option("--seed", "seed", type=int, default=2713, help="Random seed to set.")
@click.option(
    "--step-save-period",
    type=click.IntRange(0),
    help="The number of steps between intermediate model saving",
)
@click.option(
    "--strategy",
    type=click.Choice(pl_strategies.StrategyRegistry.available_strategies()),
    help="The lightning strategy to use (see lightning doc)",
)
@click.option("--profile", is_flag=True, help="Run in profiling mode")
@click.option(
    "--tf32-mode",
    type=click.Choice(["highest", "high", "medium"]),
    help="Ampere matmul optimisation mode (for supported GPUs)",
)
@click.option(
    "--tokenizer",
    "tokenizer_name",
    type=str,
    metavar="NAME_OR_PATH",
    help=(
        "A pretrained tokenizer model to use"
        " (default to the pretrained transformer model if there is one)"
    ),
)
@click.option(
    "--val-check-period",
    type=click.IntRange(0),
    help="The number of steps between validation runs (useful for very long epochs)",
)
@click.option(
    "--val-data",
    "--val-text",
    "val_path",
    help=(
        "A raw corpus for validation."
        " Either as a path or as a `handle:config:split` identifier for 🤗 hub."
        " (handle can be a url), e.g. `lgrobol/openminuscule:text:train`"
    ),
)
@click.option("--verbose", is_flag=True, help="More detailed logs")
def main(
    accelerator: str,
    cache_dir: Optional[pathlib.Path],
    checkpoint: Optional[pathlib.Path],
    config_path: Optional[pathlib.Path],
    device_batch_size: Optional[int],
    epoch_save_period: Optional[int],
    guess_batch_size: bool,
    max_epochs: Optional[int],
    max_steps: Optional[int],
    model_config_path: Optional[str],
    model_name: str,
    num_nodes: int,
    num_workers: int,
    num_devices: int,
    out_dir: pathlib.Path,
    precision: Optional[Literal["64", "32", "16", "bf16"]],
    pretrained_model: Optional[str],
    profile: bool,
    train_data: str,
    seed: int,
    step_save_period: Optional[int],
    strategy: Optional[str],
    tf32_mode: Optional[Literal["highest", "high", "medium"]],
    tokenizer_name: Optional[str],
    val_check_period: Optional[int],
    val_path: Optional[str],
    verbose: bool,
):
    """Train a Transformer model.

    The training dataset should be given with `raw_text` as either a path to a local file or or as a
    `handle:config:split` identifier for 🤗 hub (handle can be a url).
    """

    # ## Logging setup
    console = rich.get_console()
    console.stderr = True
    if (slurm_procid := os.environ.get("SLURM_PROCID")) is not None:
        log_file = out_dir / "logs" / f"train{slurm_procid}.log"
    else:
        log_file = out_dir / "train.log"
    setup_logging(
        console=console,
        log_file=log_file,
        verbose=verbose,
    )
    logger.debug(f"Current environment: {os.environ}")

    logger.info(f"Using random seed {seed}")
    pl.seed_everything(seed, workers=True)
    transformers.set_seed(seed)

    logger.debug(f"Loading config from {config_path}")
    if config_path is not None:
        with open(config_path, "rb") as in_stream:
            config = tomli.load(in_stream)
    else:
        logger.warning(
            "No config given, we'll use default values but this is probably not what you want."
        )
        config = dict()

    logger.debug("Finding out batch size depending on environment and tuning config.")
    # NOTE: this is likely duplicated somewhere in pl codebase but we need it now unless pl rolls
    # out something like `optim_batch_size` that takes into account the number of tasks and the
    # number of samples per gpu
    if (num_slurm_tasks := os.environ.get("SLURM_NTASKS")) is not None:
        total_devices = int(num_slurm_tasks)
    else:
        total_devices = num_nodes * num_devices
    logger.info(f"Training on a total of {total_devices} devices.")

    tuning_config = TrainConfig.parse_obj(config.get("tuning", dict()))

    if device_batch_size is None:
        device_batch_size = tuning_config.batch_size
    elif tuning_config.batch_size < device_batch_size * total_devices:
        raise ValueError(
            f"Batch size ({tuning_config.batch_size}) is smaller than"
            f" loader batch size({device_batch_size} samples per device × {total_devices} devices)"
            " try using fewer devices"
        )
    elif tuning_config.batch_size % (device_batch_size * total_devices):
        remainder = tuning_config.batch_size % device_batch_size * total_devices
        logger.warning(
            f"Batch size ({tuning_config.batch_size}) is not a multiple"
            f" of loader batch size({device_batch_size} samples per device × {total_devices} devices)"
            f" the actual tuning batch size used will be {tuning_config.batch_size-remainder}."
        )

    # A pl Trainer batch is in fact one batch per device, so if we use multiple devices
    accumulate_grad_batches = tuning_config.batch_size // (device_batch_size * total_devices)
    logger.info(f"Using {accumulate_grad_batches} steps gradient accumulation.")

    # In DP mode, every batch is split between the devices
    if accelerator == "dp":
        loader_batch_size = device_batch_size * total_devices
    else:
        loader_batch_size = device_batch_size

    logger.debug("Loading tokenizer, model and dataset.")
    task_type = config.get("type", "mlm")

    # TODO: make `task` here an object that will hold the config and be responsible for the rest
    task: ModuleType
    if task_type == "mlm":
        task = mlm
    elif task_type == "rtd":
        task = rtd
    elif task_type == "mbart":
        task = mbart
    else:
        raise ValueError(f"Unknown task type: {task_type!r}")

    if tokenizer_name is None:
        if pretrained_model is not None:
            tokenizer_name = pretrained_model
        else:
            raise ValueError("Missing both pretrained tokenizer and pretrained model")
    logger.info(f"Loading pretrained tokenizer {tokenizer_name}")
    tokenizer: Union[
        transformers.PreTrainedTokenizer, transformers.PreTrainedTokenizerFast
    ] = transformers.AutoTokenizer.from_pretrained(tokenizer_name, use_fast=True)

    training_model: TrainingModule = task.get_training_model(
        model_config=model_config_path,
        pretrained_model=pretrained_model,
        task_config=config.get("task"),
        tokenizer=tokenizer,
        training_config=tuning_config,
    )
    training_model.train()

    logger.info("Creating data modules")
    datamodule = training_model.get_data_module(
        data_dir=cache_dir,
        loader_batch_size=loader_batch_size,
        num_workers=num_workers,
        tokenizer=tokenizer,
        tokenizer_name=tokenizer_name.replace("/", "_"),
        train_path=train_data,
        val_path=val_path,
    )
    datamodule.prepare_data_per_node = False

    logger.info("Creating trainer")
    additional_kwargs: Dict[str, Any] = dict()
    if profile:
        logger.info("Running in profile mode")
        profiler = pl_profilers.AdvancedProfiler(dirpath=out_dir, filename="profile.txt")
        additional_kwargs.update({"profiler": profiler, "overfit_batches": 128})

    if guess_batch_size:
        logger.info("Automatic batch size selection")
        additional_kwargs.update({"auto_scale_batch_size": "binsearch"})

    if precision is not None:
        logger.info(f"Training the model using {precision} precision")
        additional_kwargs["precision"] = precision
    if accelerator == "gpu":
        if tf32_mode is not None:
            logger.info(f"Using Ampere matmul optimisations level {tf32_mode}")
            torch.set_float32_matmul_precision(tf32_mode)
    elif accelerator == "cpu":
        logger.info(f"Training the model on CPU in {num_devices} processes")
    else:
        logger.info(f"Training the model on {num_devices} devices")

    callbacks: List[pl.Callback] = [
        pl_callbacks.RichProgressBar(console_kwargs={"stderr": True}),
        pl_callbacks.LearningRateMonitor("step"),
    ]
    if epoch_save_period is not None or step_save_period is not None:
        training_model.save_transformer(out_dir / "partway_models" / "initial", tokenizer)
        callbacks.append(
            SavePretrainedModelCallback(
                out_dir / "partway_models",
                tokenizer,
                epoch_period=epoch_save_period,
                step_period=step_save_period,
            )
        )

    if checkpoint is not None:
        logger.info(f"Restarting training from the checkpoint at {checkpoint}")
        additional_kwargs["resume_from_checkpoint"] = checkpoint

    if max_steps is None:
        if tuning_config.lr_decay_steps is not None and tuning_config.lr_decay_steps != -1:
            max_steps = tuning_config.lr_decay_steps + tuning_config.warmup_steps
            logger.info(
                f"Setting the max number of steps at {max_steps} according to the tuning config"
            )
        else:
            max_steps = -1

    if val_check_period is not None:
        additional_kwargs["val_check_interval"] = val_check_period

    # TODO(2023-02-08): re-evaluate to see if we can deactivate `find_unused_parameters` again
    trainer = pl.Trainer(
        accumulate_grad_batches=accumulate_grad_batches,
        accelerator=accelerator,
        callbacks=callbacks,
        default_root_dir=str(out_dir),
        devices=num_devices,
        gradient_clip_val=tuning_config.gradient_clipping,
        limit_val_batches=1.0 if val_path is not None else 0,
        max_epochs=max_epochs,
        max_steps=max_steps,
        num_nodes=num_nodes,
        strategy=strategy,
        **additional_kwargs,
    )

    logger.info("Start training")

    trainer.fit(training_model, datamodule=datamodule)

    save_dir = out_dir / model_name
    logger.info(f"Saving the final model in {save_dir}")
    training_model.save_transformer(save_dir, tokenizer)


if __name__ == "__main__":
    main()
